"""A TestCase that initializes the library with standard API methods."""

from collections.abc import Iterable
import contextlib
import copy
import json
import os
from typing import Any

from googleapiclient import discovery

import unittest
import ee
from ee import _cloud_api_utils
from ee import _state


# Cached algorithms list
_algorithms_cache: dict[str, Any] | None = None


def GetAlgorithms() -> dict[str, Any]:
  """Returns a static version of the ListAlgorithms call.

  After ApiTestCase.setUp is called, ee.data.getAlgorithms() is patched to use
  this data. This function may be called explicitly if the test data is needed
  at a time before the test case has started executing.
  """
  global _algorithms_cache
  if _algorithms_cache is not None:
    return _algorithms_cache
  algorithms_path = os.path.join(
     os.path.dirname(os.path.realpath(__file__)),
     "tests/algorithms.json")
  with open(algorithms_path, encoding='utf-8') as algorithms_file:
    algorithms = json.load(algorithms_file)
    _algorithms_cache = _cloud_api_utils.convert_algorithms(algorithms)
  return _algorithms_cache


class ApiTestCase(unittest.TestCase):
  """A TestCase that initializes the library with standard API methods."""
  last_download_call: Any | None
  last_thumb_call: Any | None
  last_table_call: Any | None
  last_mapid_call: Any | None

  def setUp(self):
    super().setUp()
    # Store the state of ee.data before it is overwritten. We need to restore to
    # this state in tearDown to prevent state leakage between tests.
    self.old_get_algorithms = ee.data.getAlgorithms
    self.old_compute_value = ee.data.computeValue
    self.old_get_map_id = ee.data.getMapId
    self.old_get_download_id = ee.data.getDownloadId
    self.old_get_thumb_id = ee.data.getThumbId
    self.old_get_table_download_id = ee.data.getTableDownloadId
    # pylint: disable-next=protected-access
    self.old_fetch_data_catalog_stac = ee.deprecation._FetchDataCatalogStac
    self.InitializeApi()

  def tearDown(self):
    super().tearDown()
    ee.data.getAlgorithms = self.old_get_algorithms
    ee.data.computeValue = self.old_compute_value
    ee.data.getMapId = self.old_get_map_id
    ee.data.getDownloadId = self.old_get_download_id
    ee.data.getThumbId = self.old_get_thumb_id
    ee.data.getTableDownloadId = self.old_get_table_download_id
    _state.reset_state()
    # pylint: disable-next=protected-access
    ee.deprecation._FetchDataCatalogStac = self.old_fetch_data_catalog_stac

  def InitializeApi(self):
    """Initializes the library with standard API methods.

    This is normally invoked during setUp(), but subclasses may invoke
    it manually instead if they prefer.
    """
    self.last_download_call = None
    self.last_thumb_call = None
    self.last_table_call = None
    self.last_mapid_call = None

    ee.Reset()

    ee.data._install_cloud_api_resource = lambda: None  # pylint: disable=protected-access
    ee.data.getAlgorithms = GetAlgorithms
    ee.data.computeValue = lambda x: {'value': 'fakeValue'}
    ee.data.getMapId = self._MockMapId
    ee.data.getDownloadId = self._MockDownloadUrl
    ee.data.getTableDownloadId = self._MockTableDownload
    # pylint: disable-next=protected-access
    ee.deprecation._FetchDataCatalogStac = self._MockFetchDataCatalogStac
    ee.Initialize(None, '', project='my-project')

  # We are mocking the url here so the unit tests are happy.
  def _MockMapId(self, params: dict[str, Any]) -> dict[str, str]:
    self.last_mapid_call = {'url': '/mapid', 'data': params}
    return {'mapid': 'fakeMapId', 'token': 'fakeToken'}

  def _MockDownloadUrl(self, params: dict[str, Any]) -> dict[str, str]:
    self.last_download_call = {'url': '/download', 'data': params}
    return {'docid': '1', 'token': '2'}

  def _MockTableDownload(self, params: dict[str, Any]) -> dict[str, str]:
    self.last_table_call = {'url': '/table', 'data': params}
    return {'docid': '5', 'token': '6'}

  def _MockFetchDataCatalogStac(self) -> dict[str, Any]:
    return {}


def _GenerateCloudApiResource(mock_http: Any, raw: Any) -> discovery.Resource:
  """Returns a Cloud API resource for testing."""
  discovery_doc_path = os.path.join(
     os.path.dirname(os.path.realpath(__file__)),
     "tests/cloud_api_discovery_document.json")
  with open(discovery_doc_path) as discovery_doc_file:
    discovery_doc_str = discovery_doc_file.read()
  return _cloud_api_utils.build_cloud_resource_from_document(
      json.loads(discovery_doc_str),
      http_transport=mock_http,
      headers_supplier=ee.data._make_request_headers,  # pylint: disable=protected-access
      response_inspector=ee.data._handle_profiling_response,  # pylint: disable=protected-access
      raw=raw,
  )


@contextlib.contextmanager  # pytype: disable=wrong-arg-types
def UsingCloudApi(
    cloud_api_resource: Any | None = None,
    cloud_api_resource_raw: Any | None = None,
    mock_http: Any | None = None,
) -> Iterable[Any]:  # pytype: disable=wrong-arg-types
  """Returns a context manager under which the Cloud API is enabled."""
  # pylint: disable=protected-access
  old_state = copy.copy(_state._state)
  # pylint: enable=protected-access
  try:
    if cloud_api_resource is None:
      cloud_api_resource = _GenerateCloudApiResource(mock_http, False)
    if cloud_api_resource_raw is None:
      cloud_api_resource_raw = _GenerateCloudApiResource(mock_http, True)
    state = _state.get_state()
    state.cloud_api_resource = cloud_api_resource
    state.cloud_api_resource_raw = cloud_api_resource_raw
    state.initialized = True
    yield
  finally:
    # pylint: disable=protected-access
    _state._state = old_state
    # pylint: enable=protected-access


# A sample of encoded EE API JSON, used by SerializerTest and DeserializerTest.
ENCODED_JSON_SAMPLE = {
    'type': 'CompoundValue',
    'scope': [
        ['0', {
            'type': 'Invocation',
            'functionName': 'Date',
            'arguments': {
                'value': 1234567890000
            }
        }],
        ['1', {
            'type': 'LineString',
            'coordinates': [[1, 2], [3, 4]],
            'crs': {
                'type': 'name',
                'properties': {
                    'name': 'SR-ORG:6974'
                }
            }
        }],
        ['2', {
            'evenOdd': True,
            'type': 'Polygon',
            'coordinates': [
                [[0, 0], [10, 0], [10, 10], [0, 10], [0, 0]],
                [[5, 6], [7, 6], [7, 8], [5, 8]],
                [[1, 1], [2, 1], [2, 2], [1, 2]]
            ]
        }],
        ['3', {
            'type': 'Bytes',
            'value': 'aGVsbG8='
        }],
        ['4', {
            'type': 'Invocation',
            'functionName': 'String.cat',
            'arguments': {
                'string1': 'x',
                'string2': 'y'
            }
        }],
        ['5', {
            'type': 'Dictionary',
            'value': {
                'foo': 'bar',
                'baz': {'type': 'ValueRef', 'value': '4'}
            }
        }],
        ['6', {
            'type': 'Function',
            'argumentNames': ['x', 'y'],
            'body': {'type': 'ArgumentRef', 'value': 'y'}
        }],
        ['7', [
            None,
            True,
            5,
            7,
            3.4,
            112233445566778899,
            'hello',
            {'type': 'ValueRef', 'value': '0'},
            {'type': 'ValueRef', 'value': '1'},
            {'type': 'ValueRef', 'value': '2'},
            {'type': 'ValueRef', 'value': '3'},
            {'type': 'ValueRef', 'value': '5'},
            {'type': 'ValueRef', 'value': '4'},
            {'type': 'ValueRef', 'value': '6'}
        ]]
    ],
    'value': {'type': 'ValueRef', 'value': '7'}
}

# A sample of encoded EE API JSON for the Cloud API, used by SerializerTest.
ENCODED_CLOUD_API_JSON_SAMPLE = {
    'values': {
        '0': {
            'arrayValue': {
                'values': [
                    {'constantValue': None},
                    {'constantValue': True},
                    {'constantValue': 5},
                    {'constantValue': 7},
                    {'constantValue': 3.4},
                    {'integerValue': '112233445566778899'},
                    {'constantValue': 'hello'},
                    {'functionInvocationValue': {
                        'functionName': 'Date',
                        'arguments': {'value': {'constantValue': 1234567890000}}
                    }},
                    {'functionInvocationValue': {
                        'functionName': 'GeometryConstructors.LineString',
                        'arguments': {
                            'crs': {'functionInvocationValue': {
                                'functionName': 'Projection',
                                'arguments': {
                                    'crs': {'constantValue': 'SR-ORG:6974'}}
                            }},
                            'coordinates': {'arrayValue': {'values': [
                                {'valueReference': '1'},
                                {'constantValue': [3, 4]}
                            ]}}
                        }}},
                    {'functionInvocationValue': {
                        'functionName': 'GeometryConstructors.Polygon',
                        'arguments': {
                            'coordinates': {'arrayValue': {'values': [
                                {'arrayValue': {'values': [
                                    {'valueReference': '2'},
                                    {'constantValue': [10, 0]},
                                    {'constantValue': [10, 10]},
                                    {'constantValue': [0, 10]},
                                    {'valueReference': '2'}]}},
                                {'constantValue':
                                 [[5, 6], [7, 6], [7, 8], [5, 8]]},
                                {'arrayValue': {'values': [
                                    {'constantValue': [1, 1]},
                                    {'constantValue': [2, 1]},
                                    {'constantValue': [2, 2]},
                                    {'valueReference': '1'}]}}
                            ]}},
                            'evenOdd': {'constantValue': True}}}},
                    {'bytesValue': 'aGVsbG8='},
                    {'dictionaryValue': {
                        'values': {
                            'baz': {'valueReference': '3'},
                            'foo': {'constantValue': 'bar'},
                        }
                    }},
                    {'valueReference': '3'},
                    {'functionDefinitionValue': {
                        'argumentNames': ['x', 'y'],
                        'body': '4'}
                    }
                ]}},
        '1': {'constantValue': [1, 2]},
        '2': {'constantValue': [0, 0]},
        '3': {'functionInvocationValue': {
            'functionName': 'String.cat',
            'arguments': {
                'string1': {'constantValue': 'x'},
                'string2': {'constantValue': 'y'}
            }}},
        '4': {'argumentReference': 'y'},
    },
    'result': '0'
}
ENCODED_CLOUD_API_JSON_SAMPLE_PRETTY = {
    'arrayValue': {
        'values': [
            {'constantValue': None},
            {'constantValue': True},
            {'constantValue': 5},
            {'constantValue': 7},
            {'constantValue': 3.4},
            {'integerValue': '112233445566778899'},
            {'constantValue': 'hello'},
            {'functionInvocationValue': {
                'functionName': 'Date',
                'arguments': {'value': {'constantValue': 1234567890000}}
            }},
            {'functionInvocationValue': {
                'functionName': 'GeometryConstructors.LineString',
                'arguments': {
                    'crs': {'functionInvocationValue': {
                        'functionName': 'Projection',
                        'arguments': {
                            'crs': {'constantValue': 'SR-ORG:6974'}}
                    }},
                    'coordinates': {'constantValue': [[1, 2], [3, 4]]}
                }}},
            {'functionInvocationValue': {
                'functionName': 'GeometryConstructors.Polygon',
                'arguments': {
                    'coordinates': {
                        'constantValue':
                        [[[0, 0], [10, 0], [10, 10], [0, 10], [0, 0]],
                         [[5, 6], [7, 6], [7, 8], [5, 8]],
                         [[1, 1], [2, 1], [2, 2], [1, 2]]]},
                    'evenOdd': {'constantValue': True}}}},
            {'bytesValue': 'aGVsbG8='},
            {'dictionaryValue': {
                'values': {
                    'baz': {'functionInvocationValue': {
                        'functionName': 'String.cat',
                        'arguments': {
                            'string1': {'constantValue': 'x'},
                            'string2': {'constantValue': 'y'}
                        }}},
                    'foo': {'constantValue': 'bar'},
                }}},
            {'functionInvocationValue': {
                'functionName': 'String.cat',
                'arguments': {
                    'string1': {'constantValue': 'x'},
                    'string2': {'constantValue': 'y'}
                }}},
            {'functionDefinitionValue': {
                'argumentNames': ['x', 'y'],
                'body': {'argumentReference': 'y'}}
            }
        ]}}



#!/usr/bin/env python3
"""Test for the ee.dictionary module."""
import json
from typing import Any
import unittest

import unittest
import ee
from ee import apitestcase


def make_expression_graph(
    function_invocation_value: dict[str, Any],
) -> dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class DictionaryTest(apitestcase.ApiTestCase):

  def test_dictionary(self):
    """Verifies basic behavior of ee.Dictionary."""
    src = {'a': 1, 'b': 2, 'c': 'three'}
    dictionary = ee.Dictionary(src)
    self.assertEqual(
        {'type': 'Dictionary', 'value': src},
        ee.Serializer(False)._encode(dictionary),
    )
    self.assertEqual(
        {'constantValue': {'a': 1, 'b': 2, 'c': 'three'}},
        ee.Serializer(False, for_cloud_api=True)._encode(dictionary),
    )

    f = ee.Feature(None, {'properties': src})
    computed = ee.Dictionary(f.get('properties'))
    self.assertIsInstance(computed, ee.Dictionary)

    # The 4 types of arguments we expect
    cons = (ee.Dictionary(src),
            ee.Dictionary(f.get('properties')),
            ee.Dictionary(),
            ee.Dictionary(('one', 1)))

    for d in cons:
      self.assertIsInstance(d, ee.ComputedObject)

  def test_internals(self):
    """Test eq(), ne() and hash()."""
    a = ee.Dictionary({'one': 1})
    b = ee.Dictionary({'two': 2})
    c = ee.Dictionary({'one': 1})

    self.assertEqual(a, a)
    self.assertNotEqual(a, b)
    self.assertEqual(a, c)
    self.assertNotEqual(b, c)
    self.assertNotEqual(hash(a), hash(b))

  def test_combine(self):
    expect = make_expression_graph({
        'arguments': {
            'first': {'constantValue': {'a': 1}},
            'second': {'constantValue': {'b': 2}},
            'overwrite': {'constantValue': True},
        },
        'functionName': 'Dictionary.combine',
    })
    expression = ee.Dictionary({'a': 1}).combine({'b': 2}, True)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).combine(
        second={'b': 2}, overwrite=True
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_contains(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'a key'},
        },
        'functionName': 'Dictionary.contains',
    })
    expression = ee.Dictionary({'a': 1}).contains('a key')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).contains(key='a key')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_from_lists(self):
    expect = make_expression_graph({
        'arguments': {
            'keys': {'constantValue': ['a']},
            'values': {'constantValue': [1]},
        },
        'functionName': 'Dictionary.fromLists',
    })
    expression = ee.Dictionary().fromLists(['a'], [1])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary().fromLists(keys=['a'], values=[1])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
            'defaultValue': {'constantValue': 'a default'},
        },
        'functionName': 'Dictionary.get',
    })
    expression = ee.Dictionary({'a': 1}).get('b', 'a default')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).get(key='b', defaultValue='a default')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get_array(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
        },
        'functionName': 'Dictionary.getArray',
    })
    expression = ee.Dictionary({'a': 1}).getArray('b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).getArray(key='b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get_geometry(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
        },
        'functionName': 'Dictionary.getGeometry',
    })
    expression = ee.Dictionary({'a': 1}).getGeometry('b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).getGeometry(key='b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get_number(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
        },
        'functionName': 'Dictionary.getNumber',
    })
    expression = ee.Dictionary({'a': 1}).getNumber('b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).getNumber(key='b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get_string(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
        },
        'functionName': 'Dictionary.getString',
    })
    expression = ee.Dictionary({'a': 1}).getString('b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).getString(key='b')
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_keys(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
        },
        'functionName': 'Dictionary.keys',
    })
    expression = ee.Dictionary({'a': 1}).keys()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).keys()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  # TODO: test_map

  def test_remove(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'selectors': {'constantValue': ['b']},
            'ignoreMissing': {'constantValue': True},
        },
        'functionName': 'Dictionary.remove',
    })
    expression = ee.Dictionary({'a': 1}).remove(['b'], True)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).remove(
        selectors=['b'], ignoreMissing=True
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_rename(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'from': {'constantValue': ['b']},
            'to': {'constantValue': ['c']},
            'overwrite': {'constantValue': True},
        },
        'functionName': 'Dictionary.rename',
    })
    expression = ee.Dictionary({'a': 1}).rename(['b'], ['c'], True)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    # Cannot use `from` kwarg as it is a python keyword.
    expression = ee.Dictionary({'a': 1}).rename(['b'], to=['c'], overwrite=True)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_select(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'selectors': {'constantValue': ['b']},
            'ignoreMissing': {'constantValue': True},
        },
        'functionName': 'Dictionary.select',
    })
    expression = ee.Dictionary({'a': 1}).select(['b'], True)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).select(
        selectors=['b'], ignoreMissing=True
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_set(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'key': {'constantValue': 'b'},
            'value': {'constantValue': 2},
        },
        'functionName': 'Dictionary.set',
    })
    expression = ee.Dictionary({'a': 1}).set('b', 2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).set(key='b', value=2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_size(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
        },
        'functionName': 'Dictionary.size',
    })
    expression = ee.Dictionary({'a': 1}).size()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_to_array(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'keys': {'constantValue': ['b']},
            'axis': {'constantValue': 2},
        },
        'functionName': 'Dictionary.toArray',
    })
    expression = ee.Dictionary({'a': 1}).toArray(['b'], 2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).toArray(keys=['b'], axis=2)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_values(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'keys': {'constantValue': ['b']},
        },
        'functionName': 'Dictionary.values',
    })
    expression = ee.Dictionary({'a': 1}).values(['b'])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).values(keys=['b'])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_to_image(self):
    expect = make_expression_graph({
        'arguments': {
            'dictionary': {'constantValue': {'a': 1}},
            'names': {'constantValue': ['b']},
        },
        'functionName': 'Dictionary.toImage',
    })
    expression = ee.Dictionary({'a': 1}).toImage(['b'])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Dictionary({'a': 1}).toImage(names=['b'])
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_encode_opt_params(self):
    a_dict = ee.Dictionary({'x': 1})

    mock_encoder = unittest.mock.Mock(return_value='encoded-value')
    result = a_dict.encode(opt_encoder=mock_encoder)

    mock_encoder.assert_called()
    self.assertEqual('encoded-value', result)

  def test_encode_cloud_value_opt_params(self):
    a_dict = ee.Dictionary({'x': 1})

    mock_encoder = unittest.mock.Mock(return_value='encoded-value')
    result = a_dict.encode_cloud_value(opt_encoder=mock_encoder)

    mock_encoder.assert_called()
    self.assertEqual({'valueReference': 'encoded-value'}, result)


if __name__ == '__main__':
  unittest.main()



"""Stores the current state of the EE library."""

from __future__ import annotations

import dataclasses
import re
from typing import Any

# Rename to avoid redefined-outer-name warning.
from google.oauth2 import credentials as credentials_lib
import requests


# The default project to use for Cloud API calls.
DEFAULT_CLOUD_API_USER_PROJECT = 'earthengine-legacy'


@dataclasses.dataclass()
class _WorkloadTag:
  """A helper class to manage the workload tag."""

  _tag: str
  _default: str

  def __init__(self):
    self._tag = ''
    self._default = ''

  def get(self) -> str:
    return self._tag

  def set(self, tag: int | str | None) -> None:
    self._tag = self.validate(tag)

  def set_default(self, new_default: int | str | None) -> None:
    self._default = self.validate(new_default)

  def reset(self) -> None:
    self._tag = self._default

  def validate(self, tag: int | str | None) -> str:
    """Returns the validated tag or raises a ValueError.

    Args:
      tag: the tag to validate.

    Raises:
      ValueError if the tag does not match the expected format.
    """
    if not tag and tag != 0:
      return ''
    tag = str(tag)
    if not re.fullmatch(r'([a-z0-9]|[a-z0-9][-_a-z0-9]{0,61}[a-z0-9])', tag):
      validation_message = (
          'Tags must be 1-63 characters, '
          'beginning and ending with a lowercase alphanumeric character '
          '([a-z0-9]) with dashes (-), underscores (_), '
          'and lowercase alphanumerics between.'
      )
      raise ValueError(f'Invalid tag, "{tag}". {validation_message}')
    return tag


@dataclasses.dataclass()
class EEState:
  """Holds the configuration and initialized resources for an EE session."""

  initialized: bool = False
  credentials: credentials_lib.Credentials | None = None
  api_base_url: str | None = None
  tile_base_url: str | None = None
  cloud_api_base_url: str | None = None
  cloud_api_key: str | None = None
  requests_session: requests.Session | None = None
  cloud_api_resource: Any | None = None
  cloud_api_resource_raw: Any | None = None
  cloud_api_user_project: str | None = DEFAULT_CLOUD_API_USER_PROJECT
  cloud_api_client_version: str | None = None
  http_transport: Any | None = None
  deadline_ms: int = 0
  max_retries: int = 5
  user_agent: str | None = None
  workload_tag: _WorkloadTag = dataclasses.field(default_factory=_WorkloadTag)


_state: EEState | None = None


def get_state() -> EEState:
  """Retrieves the EEState for the current execution context.

  Returns:
    The EEState for the current execution context.
  """

  global _state
  if not _state:
    _state = EEState()
  return _state


def reset_state() -> None:
  """Resets the EEState to the default state."""
  global _state
  _state = None



"""Converters used in the image data fetching methods."""

import io
from typing import Any


class ImageConverter:
  """Base class for image data conversion."""

  def expected_data_format(self) -> str:
    raise NotImplementedError()

  def do_conversion(self, data: bytes) -> Any:
    raise NotImplementedError()


class IdentityImageConverter(ImageConverter):
  """Data converter that applies an identity function."""

  data_format: str

  def __init__(self, data_format: str):
    self.data_format = data_format

  def expected_data_format(self) -> str:
    return self.data_format

  def do_conversion(self, data: bytes) -> bytes:
    return data


class NumPyConverter(ImageConverter):
  """Data converter that converts from a binary string to a NumPy array."""

  def expected_data_format(self) -> str:
    return 'NPY'

  def do_conversion(self, data: bytes) -> Any:
    try:
      import numpy  # pylint: disable=g-import-not-at-top
    except ImportError as exc:
      raise ImportError('Using format NUMPY_NDARRAY requires numpy.') from exc
    return numpy.load(io.BytesIO(data))


_PIXEL_DATA_CONVERTERS: dict[str, type[ImageConverter]] = {
    'NUMPY_NDARRAY': NumPyConverter
}


def from_file_format(
    file_format: str | ImageConverter,
) -> ImageConverter | None:
  if isinstance(file_format, ImageConverter):
    return file_format
  if file_format in _PIXEL_DATA_CONVERTERS:
    return _PIXEL_DATA_CONVERTERS[file_format.upper()]()
  return None



#!/usr/bin/env python3
"""Test for the ee.geometry module.

The API objects, like ee.Geometry, are usually tested by comparing the
serialized object to a golden. To be brief, the tests in this file just validate
the Geometry function's arguments instead of the serialized object.
"""

import math

from absl.testing import parameterized

import unittest
import ee
from ee import apitestcase


_EPSILON = 1e-5


class GeometryTest(apitestcase.ApiTestCase, parameterized.TestCase):

  def epsg4326_rectangle(self, west, south, east, north):
    # If we call ee.Geometry.Rectangle with geodesic=False we would get a
    # computed call.
    return ee.Geometry({
        'coordinates': [
            [[west, north], [west, south], [east, south], [east, north]]
        ],
        'type': 'Polygon',
        'geodesic': False,
    })

  @parameterized.named_parameters(
      ('_simple', (-10, -20, 10, 20), (-10, -20, 10, 20)),
      (
          '_latitude_widerThanPolesClamped_south',
          (-10, -90, 10, 73),
          (-10, -1000, 10, 73),
      ),
      (
          '_latitude_widerThanPolesClamped_north',
          (-10, -34, 10, 90),
          (-10, -34, 10, 10000),
      ),
      ('_latitude_zeroSpan', (-10, 20, 10, 20), (-10, 20, 10, 20)),
      (
          '_longitude_crossingMeridianWithOppositeSigns',
          (170, -20, 190, 20),
          (170, -20, -170, 20),
      ),
      (
          '_longitude_crossingMeridianWithNegativeSigns',
          (170, -20, 190, 20),
          (-190, -20, -170, 20),
      ),
      (
          '_longitude_crossingMeridianWithPositiveSigns',
          (170, -20, 190, 20),
          (170, -20, 190, 20),
      ),
      ('_longitude_exactlyGlobal', (-180, -20, 180, 20), (-180, -20, 180, 20)),
      (
          '_longitude_excessOfGlobalIsClamped',
          (-180, -20, 180, 20),
          (-180 - _EPSILON, -20, 180 + _EPSILON, 20),
      ),
      ('_longitude_zeroSpan', (10, -20, 10, 20), (10, -20, 10, 20)),
      (
          '_infinity_validDirection_isClamped_west',
          (-180, -20, 180, 20),
          (-math.inf, -20, 10, 20),
      ),
      (
          '_infinity_validDirection_isClamped_east',
          (-180, -20, 180, 20),
          (-10, -20, math.inf, 20),
      ),
      (
          '_infinity_validDirection_isClamped_south',
          (-10, -90, 10, 20),
          (-10, -math.inf, 10, 20),
      ),
      (
          '_infinity_validDirection_isClamped_north',
          (-10, -20, 10, 90),
          (-10, -20, 10, math.inf),
      ),
  )
  def test_bbox(self, expected, coords):
    self.assertEqual(
        self.epsg4326_rectangle(*expected),
        ee.Geometry.BBox(*coords),
    )

  def test_bbox_computed_args(self):
    ten = ee.Number(5).add(5)
    box = ee.Geometry.BBox(-10, -20, ten, 20)
    func = ee.ApiFunction.lookup('GeometryConstructors.BBox')

    self.assertIsInstance(box, ee.Geometry)
    self.assertEqual(func, box.func)
    expected_args = {
        'west': ee.Number(-10),
        'south': ee.Number(-20),
        'east': ten,
        'north': ee.Number(20),
    }
    self.assertEqual(expected_args, box.args)

  @parameterized.named_parameters(
      # Reject cases which, if we clamped them instead, would move a box whose
      # bounds lie past a pole to being a point at the pole.
      (
          '_latitude_notBeyondPoles_north',
          (-10, -100, 10, -95),
          r'Geometry\.BBox: north must be at least -90°, but was -95°',
      ),
      (
          '_latitude_notBeyondPoles_south',
          (-10, 95, 10, 100),
          r'Geometry\.BBox: south must be at most \+90°, but was 95°',
      ),
      # NaN is rejected.
      (
          '_NaN_isRejected_west',
          (math.nan, -20, 10, 20),
          r'Geometry\.BBox: west must not be nan',
      ),
      (
          '_NaN_isRejected_south',
          (-10, math.nan, 10, 20),
          r'Geometry\.BBox: south must be at most \+90°, but was nan°',
      ),
      (
          '_NaN_isRejected_east',
          (-10, -20, math.nan, 20),
          r'Geometry\.BBox: east must not be nan',
      ),
      (
          '_NaN_isRejected_north',
          (-10, -20, 10, math.nan),
          r'Geometry.BBox: north must be at least -90°, but was nan°',
      ),
      # Infinities in the wrong direction are rejected.
      (
          '_infinity_invalidDirection_isRejected_west',
          (math.inf, -20, 10, 20),
          r'Geometry\.BBox: west must not be inf',
      ),
      (
          '_infinity_invalidDirection_isRejected_south',
          (-10, math.inf, 10, 20),
          r'Geometry\.BBox: south must be at most \+90°, but was inf°',
      ),
      (
          '_infinity_invalidDirection_isRejected_east',
          (-10, -20, -math.inf, 20),
          r'Geometry\.BBox: east must not be -inf',
      ),
      (
          '_infinity_invalidDirection_isRejected_north',
          (-10, -20, 10, -math.inf),
          r'Geometry\.BBox: north must be at least -90°, but was -inf°',
      ),
  )
  def test_bbox_invalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.BBox(*coords)

  def test_line_string(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.LineString')

    geometry = ee.Geometry.LineString(coords, proj, True, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': True,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def test_line_string_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.LineString')

    geometry = ee.Geometry.LineString(coords, geodesic=True)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': True},
        geometry.args,
    )

  def test_line_string_computed_args(self):
    """Verifies that a LineString with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.LineString([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.LineString'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def test_line_string_invalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.LineString(*coords)

  def test_linear_ring(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.LinearRing')

    geometry = ee.Geometry.LinearRing(coords, proj, geodesic, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def test_linear_ring_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.LinearRing')
    geodesic = True

    geometry = ee.Geometry.LinearRing(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def test_linear_ring_computed_args(self):
    """Verifies that a LinearRing with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    ring = ee.Geometry.LinearRing([p1, p2])

    self.assertIsInstance(ring, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.LinearRing'), ring.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, ring.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def test_linear_ring_invalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.LinearRing(*coords)

  def test_multi_line_string(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiLineString')

    geometry = ee.Geometry.MultiLineString(coords, proj, geodesic, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def test_multi_line_string_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiLineString')
    geodesic = True

    geometry = ee.Geometry.MultiLineString(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def test_multi_line_string_computed_args(self):
    """Verifies that a MultiLineString with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.MultiLineString([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiLineString'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[[1, 2], [3, 4]]]]], 'Invalid geometry'),
  )
  def test_multi_line_string_invalid(self, coords, message):
    """Verifies MultiLineString constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiLineString(*coords)

  def test_multi_point(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPoint')

    geometry = ee.Geometry.MultiPoint(coords, proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
        },
        geometry.args,
    )

  def test_multi_point_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPoint')
    proj = 'EPSG:4326'

    geometry = ee.Geometry.MultiPoint(coords, proj=proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'crs': ee.Projection(proj)},
        geometry.args,
    )

  def test_multi_point_computed_args(self):
    """Verifies that a MultiPoint with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    point = ee.Geometry.MultiPoint([p1, p2])

    self.assertIsInstance(point, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiPoint'), point.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, point.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def test_multi_point_invalid(self, coords, message):
    """Verifies MultiPoint constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiPoint(*coords)

  def test_multi_polygon(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon')

    geometry = ee.Geometry.MultiPolygon(
        coords, proj, geodesic, max_error, even_odd
    )
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def test_multi_polygon_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon')
    geodesic = True

    geometry = ee.Geometry.MultiPolygon(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def test_multi_polygon_computed_args(self):
    """Verifies that a MultiPolygon with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.MultiPolygon([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[[[[1, 2], [3, 4], [5, 6]]]]]], 'Invalid geometry'),
      ('_nesting', [[[[[1, 2], [3, 4]], [1, 2]]]], 'Invalid geometry'),
  )
  def test_multi_polygon_invalid(self, coords, message):
    """Verifies MultiPolygon constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiPolygon(*coords)

  def test_point(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    func = ee.ApiFunction.lookup('GeometryConstructors.Point')

    geometry = ee.Geometry.Point(coords, proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
        },
        geometry.args,
    )

  def test_point_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Point')
    proj = 'EPSG:4326'

    geometry = ee.Geometry.Point(coords, proj=proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'crs': ee.Projection(proj)},
        geometry.args,
    )

  def test_point_old_keyword_args(self):
    """Verifies that Points still allow keyword lon/lat args."""
    self.assertEqual(ee.Geometry.Point(1, 2), ee.Geometry.Point(lon=1, lat=2))
    self.assertEqual(ee.Geometry.Point(1, 2), ee.Geometry.Point(1, lat=2))

  def test_point_computed_args(self):
    """Verifies that a Point with computed inputs can be constructed."""
    v1 = ee.Number(1)
    v2 = ee.Number(2)
    point = ee.Geometry.Point([v1, v2])

    self.assertIsInstance(point, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), point.func
    )
    self.assertEqual({'coordinates': ee.List([v1, v2])}, point.args)

  @parameterized.named_parameters(
      ('_geometry', [['-78.204948', '40.966539']], 'Invalid geometry'),
  )
  def test_point_invalid(self, coords, message):
    """Verifies Point constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Point(*coords)

  def test_polygon(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.Polygon')

    geometry = ee.Geometry.Polygon(coords, proj, geodesic, max_error, even_odd)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def test_polygon_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Polygon')
    geodesic = True

    geometry = ee.Geometry.Polygon(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def test_polygon_computed_args(self):
    """Verifies that a Polygon with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.Polygon([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Polygon'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  def test_polygon_even_odd(self):
    poly1 = ee.Geometry.Polygon([0, 0, 0, 5, 5, 0])
    self.assertTrue(poly1.toGeoJSON()['evenOdd'])

    # TODO: Use kwargs instead of positional args for evenOdd.
    poly2 = ee.Geometry.Polygon([0, 0, 0, 5, 5, 0], None, None, None, False)
    self.assertFalse(poly2.toGeoJSON()['evenOdd'])

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_tooMuchNesting', [[[[[1, 2], [3, 4], [5, 6]]]]], 'Invalid geometry'),
      ('_badNesting', [[[[1, 2], [3, 4]], [1, 2]]], 'Invalid geometry'),
  )
  def test_polygon_invalid(self, coords, message):
    """Verifies Polygon constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Polygon(*coords)

  def test_rectangle(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.Rectangle')

    geometry = ee.Geometry.Rectangle(coords, proj, geodesic, even_odd)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def test_rectangle_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Rectangle')
    geodesic = True

    geometry = ee.Geometry.Rectangle(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def test_rectangle_old_keyword_args(self):
    """Verifies that Rectangles still allow keyword xlo/ylo/xhi/yhi args."""
    self.assertEqual(
        ee.Geometry.Rectangle(1, 2, 3, 4),
        ee.Geometry.Rectangle(xlo=1, ylo=2, xhi=3, yhi=4),
    )
    self.assertEqual(
        ee.Geometry.Rectangle(1, 2, 3, 4),
        ee.Geometry.Rectangle(1, 2, xhi=3, yhi=4),
    )

  def test_rectangle_computed_args(self):
    """Verifies that a Rectangle with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.Rectangle([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Rectangle'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_tooMuchNesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def test_rectangle_invalid(self, coords, message):
    """Verifies Rectangle constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Rectangle(*coords)

  def test_valid_geometry_collection(self):
    """Verifies GeometryCollection constructor behavior with valid arguments."""
    geometry = ee.Geometry({
        'type': 'GeometryCollection',
        'geometries': [
            {
                'type': 'Polygon',
                'coordinates': [[[-1, -1], [0, 1], [1, -1]]],
                'geodesic': True,
                'evenOdd': True,
            },
            {'type': 'Point', 'coordinates': [0, 0]},
            {
                'type': 'GeometryCollection',
                'geometries': [
                    {'type': 'Point', 'coordinates': [1, 2]},
                    {'type': 'Point', 'coordinates': [2, 1]},
                ],
            },
        ],
        'coordinates': [],
    })
    self.assertIsInstance(geometry, ee.Geometry)

  def test_array_constructors(self):
    """Verifies that constructors that take arrays fix nesting."""
    get_coordinates_count = lambda g: len(g.toGeoJSON()['coordinates'])

    point = ee.Geometry.Point([1, 2])
    self.assertEqual(2, get_coordinates_count(point))

    multipoint = ee.Geometry.MultiPoint([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(multipoint))

    line = ee.Geometry.LineString([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(line))

    ring = ee.Geometry.LinearRing([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(ring))

    multiline = ee.Geometry.MultiLineString(
        [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]
    )
    self.assertEqual(2, get_coordinates_count(multiline))

    polygon = ee.Geometry.Polygon([[[1, 2], [3, 4], [5, 6]]])
    self.assertEqual(1, get_coordinates_count(polygon))

    mpolygon = ee.Geometry.MultiPolygon(
        [[[[1, 2], [3, 4], [5, 6]]], [[[1, 2], [3, 4], [5, 6]]]]
    )
    self.assertEqual(2, get_coordinates_count(mpolygon))

  def test_geodesic_flag(self):
    """Verifies that JSON parsing and generation preserves the geodesic flag."""
    geodesic = ee.Geometry({
        'type': 'LineString',
        'coordinates': [[1, 2], [3, 4]],
        'geodesic': True,
    })
    projected = ee.Geometry({
        'type': 'LineString',
        'coordinates': [[1, 2], [3, 4]],
        'geodesic': False,
    })
    self.assertTrue(geodesic.toGeoJSON()['geodesic'])
    self.assertFalse(projected.toGeoJSON()['geodesic'])

  def test_constructor(self):
    """Check the behavior of the Geometry constructor.

    There are 5 options:
      1) A geoJSON object.
      2) A not-computed geometry.
      3) A not-computed geometry with overrides.
      4) A computed geometry with no overrides.
      5) something to cast to geometry.
    """
    line = ee.Geometry.LineString(1, 2, 3, 4)

    # GeoJSON.
    from_json = ee.Geometry(line.toGeoJSON())
    self.assertIsNone(from_json.func)
    self.assertEqual(from_json._type, 'LineString')
    self.assertEqual(from_json._coordinates, [[1, 2], [3, 4]])

    # GeoJSON with a CRS specified.
    json_with_crs = line.toGeoJSON()
    json_with_crs['crs'] = {
        'type': 'name',
        'properties': {'name': 'SR-ORG:6974'},
    }
    from_json_with_crs = ee.Geometry(json_with_crs)
    self.assertIsNone(from_json_with_crs.func)
    self.assertEqual(from_json_with_crs._type, 'LineString')
    self.assertEqual(from_json_with_crs._proj, 'SR-ORG:6974')

    # A not-computed geometry.
    self.assertEqual(ee.Geometry(line), line)

    # A not-computed geometry with an override.
    with_override = ee.Geometry(line, 'SR-ORG:6974')
    self.assertEqual(with_override._proj, 'SR-ORG:6974')

    # A computed geometry.
    self.assertEqual(ee.Geometry(line.bounds()), line.bounds())

    # Something to cast to a geometry.
    computed = ee.ComputedObject(ee.Function(), {'a': 1})
    geom = ee.Geometry(computed)
    self.assertEqual(computed.func, geom.func)
    self.assertEqual(computed.args, geom.args)

  @parameterized.named_parameters(
      ('_proj', {'proj': 'some-proj'}),
      ('_geodesic', {'geodesic': False}),
      ('_geodesic_true', {'geodesic': True}),
      ('_evenOdd', {'evenOdd': False}),
      ('_evenOdd_true', {'evenOdd': True}),
  )
  def test_constructor_error_for_computed_geometry_with_overrides(self, kwargs):
    """Verifies that Geometry raises for computed geometries with overrides."""
    computed_geometry = ee.Geometry.Polygon(
        coords=[[[-35, -10], [-35, 10], [35, 10], [35, -10], [-35, -10]]],
        geodesic=True,
    )
    with self.assertRaisesRegex(
        ee.EEException,
        'Setting the CRS or geodesic on a computed Geometry is not supported.',
    ):
      ee.Geometry(computed_geometry, **kwargs)

  def test_computed_geometries(self):
    """Verifies the computed object behavior of the Geometry constructor."""
    line = ee.Geometry.LineString(1, 2, 3, 4)
    bounds = line.bounds()

    self.assertIsInstance(bounds, ee.Geometry)
    self.assertEqual(ee.ApiFunction.lookup('Geometry.bounds'), bounds.func)
    self.assertEqual(line, bounds.args['geometry'])
    self.assertTrue(hasattr(bounds, 'bounds'))

  def test_computed_coordinate(self):
    """Verifies that a computed coordinate produces a computed geometry."""
    coords = [1, ee.Number(1).add(1)]
    p = ee.Geometry.Point(coords)

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    self.assertEqual({'coordinates': ee.List(coords)}, p.args)

  def test_computed_list(self):
    """Verifies that a computed coordinate produces a computed geometry."""
    lst = ee.List([1, 2, 3, 4]).slice(0, 2)
    p = ee.Geometry.Point(lst)

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    self.assertEqual({'coordinates': lst}, p.args)

  def test_computed_projection(self):
    """Verifies that a geometry with a projection can be constructed."""
    p = ee.Geometry.Point([1, 2], 'epsg:4326')

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    expected_args = {
        'coordinates': ee.List([1, 2]),
        'crs': ee.ApiFunction.lookup('Projection').call('epsg:4326'),
    }
    self.assertEqual(expected_args, p.args)

  def test_invalid_crs(self):
    geo_json = ee.Geometry.LineString(1, 2, 3, 4).toGeoJSON()
    geo_json['crs'] = {'something': 'invalid-crs'}
    with self.assertRaisesRegex(ee.EEException, 'Invalid CRS declaration'):
      ee.Geometry(geo_json)

  def test_internals(self):
    """Test eq(), ne() and hash()."""
    a = ee.Geometry.Point(1, 2)
    b = ee.Geometry.Point(2, 1)
    c = ee.Geometry.Point(1, 2)

    self.assertEqual(a, a)
    self.assertNotEqual(a, b)
    self.assertEqual(a, c)
    self.assertNotEqual(b, c)
    self.assertNotEqual(hash(a), hash(b))

  def test_coordinates_to_line(self):
    # pylint: disable=protected-access
    self.assertEqual([], ee.Geometry._coordinatesToLine([]))
    self.assertEqual([1, 2], ee.Geometry._coordinatesToLine([1, 2]))
    # TODO: schwehr - Improve _coordinatesToLine typing.
    self.assertEqual(
        [[1, 2]],
        ee.Geometry._coordinatesToLine([[1, 2]]),  # pytype: disable=wrong-arg-types
    )
    self.assertEqual(
        [[1, 2], [3, 4]], ee.Geometry._coordinatesToLine([1, 2, 3, 4])
    )
    self.assertEqual(
        [[1, 2], [3, 4], [5, 6]],
        ee.Geometry._coordinatesToLine([1, 2, 3, 4, 5, 6]),
    )
    with self.assertRaisesRegex(
        ee.EEException, 'Invalid number of coordinates: 5'
    ):
      ee.Geometry._coordinatesToLine([1, 2, 3, 4, 5])
    # pylint: enable=protected-access

  def test_init_opt_params(self):
    result = ee.Geometry(
        geo_json={'type': 'Polygon', 'coordinates': [[[-2, 1]]]},
        opt_proj='abc',
        opt_geodesic=True,
        opt_evenOdd=True,
    ).serialize()
    self.assertIn('"crs": {"constantValue": "abc"}', result)
    self.assertIn('"geodesic": {"constantValue": true}', result)
    self.assertIn('"evenOdd": {"constantValue": true}', result)

  @parameterized.named_parameters(
      (
          'BBox',
          lambda: ee.Geometry.BBox(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 4], [1, 2], [3, 2], [3, 4]]],
              geodesic=False,
              type='Polygon',
          ),
      ),
      (
          'LineString',
          lambda: ee.Geometry.LineString(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='LineString'),
      ),
      (
          'LinearRing',
          lambda: ee.Geometry.LinearRing(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='LinearRing'),
      ),
      (
          'MultiLineString',
          lambda: ee.Geometry.MultiLineString(1, 2, 3, 4),
          dict(coordinates=[[[1, 2], [3, 4]]], type='MultiLineString'),
      ),
      (
          'MultiLineString_Empty',
          lambda: ee.Geometry.MultiLineString(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], type='MultiLineString'),
      ),
      (
          'MultiPoint',
          lambda: ee.Geometry.MultiPoint(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='MultiPoint'),
      ),
      (
          'MultiPoint_Empty',
          lambda: ee.Geometry.MultiPoint(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], type='MultiPoint'),
      ),
      (
          'MultiPolygon',
          lambda: ee.Geometry.MultiPolygon(1, 2, 3, 4),
          dict(
              coordinates=[[[[1, 2], [3, 4]]]],
              evenOdd=True,
              type='MultiPolygon',
          ),
      ),
      (
          'MultiPolygon_Empty',
          lambda: ee.Geometry.MultiPolygon(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], evenOdd=True, type='MultiPolygon'),
      ),
      (
          'Point',
          lambda: ee.Geometry.Point(coords=[1, 2]),
          dict(coordinates=[1, 2], type='Point'),
      ),
      (
          'Polygon',
          lambda: ee.Geometry.Polygon(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 2], [3, 4]]],
              evenOdd=True,
              type='Polygon',
          ),
      ),
      (
          'Rectangle',
          lambda: ee.Geometry.Rectangle(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 4], [1, 2], [3, 2], [3, 4]]],
              evenOdd=True,
              type='Polygon',
          ),
      ),
  )
  def test_to_geo_json(self, geometry_func, expected_geojson):
    self.assertEqual(expected_geojson, geometry_func().toGeoJSON())

  @parameterized.named_parameters(
      # BBox doesn't have any other arguments.
      ('LineString', lambda: ee.Geometry.LineString([], 'EPSG:4326')),
      ('LinearRing', lambda: ee.Geometry.LinearRing([], 'EPSG:4326')),
      ('MultiLineString', lambda: ee.Geometry.MultiLineString([], 'EPSG:4326')),
      ('MultiPoint', lambda: ee.Geometry.MultiPoint([], 'EPSG:4326')),
      ('MultiPolygon', lambda: ee.Geometry.MultiPolygon([], 'EPSG:4326')),
      ('Point', lambda: ee.Geometry.Point([], 'EPSG:4326')),
      ('Polygon', lambda: ee.Geometry.Polygon([], 'EPSG:4326')),
      ('Rectangle', lambda: ee.Geometry.Rectangle([], 'EPSG:4326')),
  )
  def test_to_geo_json_fails_on_computed(self, geometry_func):
    message = 'Cannot convert a computed geometry to GeoJSON'
    with self.assertRaisesRegex(ee.EEException, message):
      geometry_func().toGeoJSON()


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Terrain module."""

import json
from typing import Any

import unittest
import ee
from ee import apitestcase

# For an ee.Image(1).
IMAGE_1 = 1
IMAGE_EXPRESSION_1 = {
    'functionInvocationValue': {
        'functionName': 'Image.constant',
        'arguments': {'value': {'constantValue': IMAGE_1}},
    }
}


def make_expression_graph(
    function_invocation_value: dict[str, Any],
) -> dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class TerrainTest(apitestcase.ApiTestCase):

  def test_aspect(self):
    expect = make_expression_graph({
        'functionName': 'Terrain.aspect',
        'arguments': {'input': IMAGE_EXPRESSION_1},
    })
    expression = ee.Terrain.aspect(IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.aspect(input=IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_fill_minima(self):
    border_value = 2
    neighborhood = 3
    expect = make_expression_graph({
        'functionName': 'Terrain.fillMinima',
        'arguments': {
            'image': IMAGE_EXPRESSION_1,
            'borderValue': {'constantValue': border_value},
            'neighborhood': {'constantValue': neighborhood},
        },
    })
    expression = ee.Terrain.fillMinima(IMAGE_1, border_value, neighborhood)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.fillMinima(
        image=IMAGE_1, borderValue=border_value, neighborhood=neighborhood
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_hill_shadow(self):
    azimuth = 2
    zenith = 3
    neighborhood_size = 4
    hysteresis = False
    expect = make_expression_graph({
        'functionName': 'Terrain.hillShadow',
        'arguments': {
            'image': IMAGE_EXPRESSION_1,
            'azimuth': {'constantValue': azimuth},
            'zenith': {'constantValue': zenith},
            'neighborhoodSize': {'constantValue': neighborhood_size},
            'hysteresis': {'constantValue': hysteresis},
        },
    })
    expression = ee.Terrain.hillShadow(
        IMAGE_1, azimuth, zenith, neighborhood_size, hysteresis
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.hillShadow(
        image=IMAGE_1,
        azimuth=azimuth,
        zenith=zenith,
        neighborhoodSize=neighborhood_size,
        hysteresis=hysteresis,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_hillshade(self):
    azimuth = 2
    elevation = 3
    expect = make_expression_graph({
        'functionName': 'Terrain.hillshade',
        'arguments': {
            'input': IMAGE_EXPRESSION_1,
            'azimuth': {'constantValue': azimuth},
            'elevation': {'constantValue': elevation},
        },
    })
    expression = ee.Terrain.hillshade(IMAGE_1, azimuth, elevation)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.hillshade(
        input=IMAGE_1, azimuth=azimuth, elevation=elevation
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_products(self):
    expect = make_expression_graph({
        'functionName': 'Terrain.products',
        'arguments': {'input': IMAGE_EXPRESSION_1},
    })
    expression = ee.Terrain.products(IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.products(input=IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_slope(self):
    expect = make_expression_graph({
        'functionName': 'Terrain.slope',
        'arguments': {'input': IMAGE_EXPRESSION_1},
    })
    expression = ee.Terrain.slope(IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Terrain.slope(input=IMAGE_1)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_init(self):
    message = r'Terrain should not be used as an object.*'
    with self.assertRaisesRegex(RuntimeError, message):
      ee.Terrain()

  def test_name(self):
    self.assertEqual('Terrain', ee.Terrain.name())


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Test for the ee.element module."""

import datetime

import unittest
import ee
from ee import apitestcase


class ElementTestCase(apitestcase.ApiTestCase):

  def test_set(self):
    """Verifies Element.set() keyword argument interpretation."""
    image = ee.Image(1)

    # Constant dictionary.
    def AssertProperties(expected, image):
      properties = {}
      while image.func == ee.ApiFunction.lookup('Element.set'):
        key = image.args['key']
        if not isinstance(key, str):
          key = key.encode()
        properties[key] = image.args['value']
        image = image.args['object']
      self.assertEqual(ee.Image(1), image)
      self.assertEqual(expected, properties)

    AssertProperties({'foo': 'bar'}, image.set({'foo': 'bar'}))
    AssertProperties({'foo': 'bar'}, image.set({'properties': {'foo': 'bar'}}))
    AssertProperties({'properties': 5}, image.set({'properties': 5}))
    AssertProperties({'properties': {'foo': 'bar'}, 'baz': 'quux'},
                     image.set({'properties': {'foo': 'bar'}, 'baz': 'quux'}))
    AssertProperties({'foo': 'bar', 'baz': 'quux'},
                     image.set('foo', 'bar', 'baz', 'quux'))
    dt = datetime.datetime.fromtimestamp(12345)
    AssertProperties({'foo': dt}, image.set('foo', dt))

    # Computed dictionary.
    computed_arg = ee.ComputedObject(None, None, 'foo')

    def CheckMultiProperties(result):
      self.assertEqual(ee.ApiFunction.lookup('Element.setMulti'), result.func)
      self.assertEqual({
          'object': image,
          'properties': ee.Dictionary(computed_arg)
      }, result.args)
    CheckMultiProperties(image.set(computed_arg))
    CheckMultiProperties(image.set({'properties': computed_arg}))

  def test_set_one_arg_exception(self):
    with self.assertRaisesRegex(
        ee.EEException,
        r'When Element\.set\(\) is passed one argument, '
        r'it must be a dictionary\.',
    ):
      ee.Image(1).set('not a dictionary')

  def test_set_odd_args_exception(self):
    with self.assertRaisesRegex(
        ee.EEException,
        r'When Element\.set\(\) is passed multiple arguments, there '
        r'must be an even number of them\.',
    ):
      ee.Image(1).set('key1', 'value1', 'key2')

  def test_get_array(self):
    image = ee.Image(1)
    array_property = image.getArray('array_prop')
    self.assertIsInstance(array_property, ee.Array)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.getArray'), array_property.func
    )
    self.assertEqual(
        {'object': image, 'property': ee.String('array_prop')},
        array_property.args,
    )

  def test_get_number(self):
    image = ee.Image(1)
    number_property = image.getNumber('number_prop')
    self.assertIsInstance(number_property, ee.Number)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.getNumber'), number_property.func
    )
    self.assertEqual(
        {'object': image, 'property': ee.String('number_prop')},
        number_property.args,
    )

  def test_get_string(self):
    image = ee.Image(1)
    string_property = image.getString('string_prop')
    self.assertIsInstance(string_property, ee.String)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.getString'), string_property.func
    )
    self.assertEqual(
        {'object': image, 'property': ee.String('string_prop')},
        string_property.args,
    )

  def test_property_names(self):
    image = ee.Image(1)
    property_names = image.propertyNames()
    self.assertIsInstance(property_names, ee.List)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.propertyNames'), property_names.func
    )
    self.assertEqual({'element': image}, property_names.args)

  def test_to_dictionary(self):
    image = ee.Image(1)
    dictionary = image.toDictionary()
    self.assertIsInstance(dictionary, ee.Dictionary)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.toDictionary'), dictionary.func
    )
    self.assertEqual({'element': image, 'properties': None}, dictionary.args)

    dictionary_with_props = image.toDictionary(['a', 'b'])
    self.assertIsInstance(dictionary_with_props, ee.Dictionary)
    self.assertEqual(
        ee.ApiFunction.lookup('Element.toDictionary'),
        dictionary_with_props.func,
    )
    self.assertEqual(
        {'element': image, 'properties': ee.List(['a', 'b'])},
        dictionary_with_props.args,
    )

  def test_init_opt_params(self):
    result = ee.Element(func=None, args=None, opt_varName='test').serialize()
    self.assertIn('"0": {"argumentReference": "test"}', result)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Classifiers."""
from __future__ import annotations

from ee import _arg_types
from ee import apifunction
from ee import computedobject
from ee import confusionmatrix
from ee import dictionary
from ee import ee_list
from ee import ee_string


class Classifier(computedobject.ComputedObject):
  """An object to represent an Earth Engine Classifier.

  Example:
    # https://developers.google.com/earth-engine/guides/classification
  """

  _initialized: bool = False

  def __init__(self, classifier: computedobject.ComputedObject):
    """Creates a Classifier wrapper.

    Args:
      classifier: A Classifier to cast.
    """
    self.initialize()

    if isinstance(classifier, computedobject.ComputedObject):
      # There is no server-side constructor for ee.Classifier. Pass the object
      # as-is to the server in case it is intended to be a Classifier cast.
      super().__init__(classifier.func, classifier.args, classifier.varName)
      return

    raise TypeError(
        'Classifier can only be used as a cast to Classifier. Found'
        f' {type(classifier)}.'
    )

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Classifier'

  def amnhMaxent(
      # pylint: disable=invalid-name
      categoricalNames: _arg_types.List | None = None,
      outputFormat: _arg_types.String | None = None,
      autoFeature: _arg_types.Bool | None = None,
      # pylint: enable=invalid-name
      linear: _arg_types.Bool | None = None,
      quadratic: _arg_types.Bool | None = None,
      product: _arg_types.Bool | None = None,
      threshold: _arg_types.Bool | None = None,
      hinge: _arg_types.Bool | None = None,
      # pylint: disable=invalid-name
      hingeThreshold: _arg_types.Integer | None = None,
      l2lqThreshold: _arg_types.Integer | None = None,
      lq2lqptThreshold: _arg_types.Integer | None = None,
      addSamplesToBackground: _arg_types.Bool | None = None,
      addAllSamplesToBackground: _arg_types.Bool | None = None,
      betaMultiplier: _arg_types.Number | None = None,
      betaHinge: _arg_types.Number | None = None,
      betaLqp: _arg_types.Number | None = None,
      betaCategorical: _arg_types.Number | None = None,
      betaThreshold: _arg_types.Number | None = None,
      # pylint: enable=invalid-name
      extrapolate: _arg_types.Bool | None = None,
      # pylint: disable=invalid-name
      doClamp: _arg_types.Bool | None = None,
      writeClampGrid: _arg_types.Bool | None = None,
      randomTestPoints: _arg_types.Integer | None = None,
      # pylint: enable=invalid-name
      seed: _arg_types.Integer | None = None,
  ) -> Classifier:
    """Returns a Maximum Entropy classifier.

    Maxent is used to model species distribution probabilities using
    environmental data for locations of known presence and for a large number of
    'background' locations. For more information and to cite, see:
    https://biodiversityinformatics.amnh.org/open_source/maxent/ and the
    reference publication: Phillips, et al., 2004. A maximum entropy approach
    to species distribution modeling, Proceedings of the Twenty-First
    International Conference on Machine Learning. The output is a single band
    named 'probability', containing the modeled probability,  and an additional
    band named 'clamp' when the 'writeClampGrid' argument is true.

    Args:
      categoricalNames: A list of the names of the categorical inputs. Any
        inputs not listed in this argument are considered to be continuous.
      outputFormat: Representation of probabilities in output.
      autoFeature: Automatically select which feature classes to use, based on
        number of training samples.
      linear: Allow linear features to be used. Ignored when autofeature is
        true.
      quadratic: Allow quadratic features to be used. Ignored when autofeature
        is true.
      product: Allow product features to be used. Ignored when autofeature is
        true.
      threshold: Allow threshold features to be used. Ignored when autofeature
        is true.
      hinge: Allow hinge features to be used. Ignored when autofeature is true.
      hingeThreshold: Number of samples at which hinge features start being
        used. Ignored when autofeature is false.
      l2lqThreshold: Number of samples at which quadratic features start being
        used. Ignored when autofeature is false.
      lq2lqptThreshold: Number of samples at which product and threshold
        features start being used. Ignored when autofeature is false.
      addSamplesToBackground: Add to the background any sample for which has a
        combination of environmental values that isn't already present in the
        background.
      addAllSamplesToBackground: Add all samples to the background, even if they
        have combinations of environmental values that are already present in
        the background.
      betaMultiplier: Regularization multiplier. Multiply all automatic
        regularization parameters by this number. A higher number gives a more
        spread-out distribution.
      betaHinge: Regularization parameter to be applied to all hinge features;
        negative value enables automatic setting.
      betaLqp: Regularization parameter to be applied to all linear, quadratic
        and product features; negative value enables automatic setting.
      betaCategorical: Regularization parameter to be applied to all categorical
        features; negative value enables automatic setting.
      betaThreshold: Regularization parameter to be applied to all threshold
        features; negative value enables automatic setting.
      extrapolate: Extrapolate. Predict to regions of environmental space
        outside the limits encountered during training.
      doClamp: Apply clamping to output.
      writeClampGrid: Adds a band to the output ('clamp') showing the spatial
        distribution of clamping. At each point, the value is the absolute
        difference between prediction values with and without clamping.
      randomTestPoints: Random test percentage. The percentage of training
        points to hold aside as test points, used to compute AUX, omission, etc.
      seed: A seed used when generating random numbers.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.amnhMaxent',
        categoricalNames,
        outputFormat,
        autoFeature,
        linear,
        quadratic,
        product,
        threshold,
        hinge,
        hingeThreshold,
        l2lqThreshold,
        lq2lqptThreshold,
        addSamplesToBackground,
        addAllSamplesToBackground,
        betaMultiplier,
        betaHinge,
        betaLqp,
        betaCategorical,
        betaThreshold,
        extrapolate,
        doClamp,
        writeClampGrid,
        randomTestPoints,
        seed,
    )

  def confusionMatrix(self) -> confusionmatrix.ConfusionMatrix:
    """Returns a 2D confusion matrix.

    Computes a 2D confusion matrix for a classifier based on its training data
    (e.g., resubstitution error). Axis 0 of the matrix corresponds to the input
    classes and axis 1 corresponds to the output classes. The rows and columns
    start at class 0 and increase sequentially up to the maximum class value, so
    some rows or columns might be empty if the input classes aren't 0-based or
    sequential.
    """

    return apifunction.ApiFunction.call_('Classifier.confusionMatrix', self)

  @staticmethod
  def decisionTree(
      treeString: _arg_types.String,  # pylint: disable=invalid-name
  ) -> Classifier:
    """Returns a classifier that applies the given decision tree.

    Args:
      treeString: The decision tree, specified in the text format generated by R
        and other similar tools.
    """

    return apifunction.ApiFunction.call_('Classifier.decisionTree', treeString)

  @staticmethod
  def decisionTreeEnsemble(
      treeStrings: _arg_types.List,  # pylint: disable=invalid-name
  ) -> Classifier:
    """Creates a classifier that applies the given decision trees.

    Args:
      treeStrings: The decision trees, specified in the text format generated by
        R and other similar tools. Each item in the list should contain one or
        more trees in text format.

    Returns:
      An ee.Classifier.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.decisionTreeEnsemble', treeStrings
    )

  def explain(self) -> dictionary.Dictionary:
    """Returns a dictionary describing the results of a trained classifier."""

    return apifunction.ApiFunction.call_(self.name() + '.explain', self)

  @staticmethod
  def libsvm(
      # pylint: disable=invalid-name
      decisionProcedure: _arg_types.String | None = None,
      svmType: _arg_types.String | None = None,
      kernelType: _arg_types.String | None = None,
      # pylint: enable=invalid-name
      shrinking: _arg_types.Bool | None = None,
      degree: _arg_types.Integer | None = None,
      gamma: _arg_types.Number | None = None,
      coef0: _arg_types.Number | None = None,
      cost: _arg_types.Number | None = None,
      nu: _arg_types.Number | None = None,
      # pylint: disable=invalid-name
      terminationEpsilon: _arg_types.Number | None = None,
      lossEpsilon: _arg_types.Number | None = None,
      oneClass: _arg_types.Integer | None = None,
      # pylint: enable=invalid-name
  ) -> Classifier:
    """Returns an empty Support Vector Machine classifier.

    Args:
      decisionProcedure: The decision procedure to use for classification.
        Either 'Voting' or 'Margin'. Not used for regression.
      svmType: The SVM type. One of `C_SVC`, `NU_SVC`, `ONE_CLASS`,
        `EPSILON_SVR`, or `NU_SVR`.
      kernelType: The kernel type. One of LINEAR, POLY, RBF, or SIGMOID.
      shrinking: Whether to use shrinking heuristics.
      degree: The degree of polynomial. Valid for POLY kernels.
      gamma: The gamma value in the kernel function. Defaults to the reciprocal
        of the number of features. Valid for POLY, RBF, and SIGMOID kernels.
      coef0: The coef₀ value in the kernel function. Defaults to 0. Valid for
        POLY and SIGMOID kernels.
      cost: The cost (C) parameter. Defaults to 1. Only valid for C-SVC,
        epsilon-SVR, and nu-SVR.
      nu: The nu parameter. Defaults to 0.5. Only valid for nu-SVC, one-class
        SVM, and nu-SVR.
      terminationEpsilon: The termination criterion tolerance (e). Defaults to
        0.001. Only valid for epsilon-SVR.
      lossEpsilon: The epsilon in the loss function (p). Defaults to 0.1. Only
        valid for epsilon-SVR.
      oneClass: The class of the training data on which to train in a one-class
        SVM. Defaults to 0. Only valid for one-class SVM. Possible values are 0
        and 1. The classifier output is binary (0/1) and will match this class
        value for the data determined to be in the class.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.libsvm',
        decisionProcedure,
        svmType,
        kernelType,
        shrinking,
        degree,
        gamma,
        coef0,
        cost,
        nu,
        terminationEpsilon,
        lossEpsilon,
        oneClass,
    )

  @staticmethod
  # pylint: disable-next=redefined-builtin
  def load(id: _arg_types.String) -> Classifier:
    """Returns a classifier from an asset.

    Args:
      id: The Classifier's Asset ID.

    Returns:
      An ee.Classifier.
    """

    return apifunction.ApiFunction.call_('Classifier.load', id)

  @staticmethod
  def minimumDistance(
      metric: _arg_types.String | None = None,
      # pylint: disable-next=invalid-name
      kNearest: _arg_types.Integer | None = None,
  ) -> Classifier:
    # pyformat: disable
    """Returns a minimum distance classifier for the given distance metric.

    Creates a minimum distance classifier for the given distance metric. In
    CLASSIFICATION mode, the nearest class is returned. In REGRESSION mode, the
    distance to the nearest class center is returned. In RAW mode, the distance
    to every class center is returned.

    Args:
      metric: The distance metric to use. Options are:
        * 'euclidean' - Euclidean distance from the unnormalized class mean.
        * 'cosine' - spectral angle from the unnormalized class mean.
        * 'mahalanobis' - Mahalanobis distance from the class mean.
        * 'manhattan' - Manhattan distance from the unnormalized class mean.
      kNearest: If greater than 1, the result will contain an array of the k
        nearest neighbors or distances, based on the output mode setting. If
        kNearest is greater than the total number of classes, it will be set
        equal to the number of classes.
    """
    # pyformat: enable

    return apifunction.ApiFunction.call_(
        'Classifier.minimumDistance', metric, kNearest
    )

  def mode(self) -> ee_string.String:
    """Returns the classifier mode string.

    The string will be one of `CLASSIFICATION`, `REGRESSION`, `PROBABILITY`,
    `MULTIPROBABILITY`, `RAW`, or `RAW_REGRESSION`.
    """

    return apifunction.ApiFunction.call_(self.name() + '.mode', self)

  def schema(self) -> ee_list.List:
    """Returns a list of the schema of the classifier.

    Returns the names of the inputs used by this classifier or null if this
    classifier has not had any training data added yet.
    """

    return apifunction.ApiFunction.call_(self.name() + '.schema', self)

  def setOutputMode(self, mode: _arg_types.String) -> Classifier:
    # pyformat: disable
    """Returns a classifier with the given output mode.

    Refer to https://developers.google.com/earth-engine/guides/classification
    for a list of supported modes for each classifier.

    Args:
      mode: The output mode. One of:
        * CLASSIFICATION (default): The output is the class number.
        * REGRESSION: The output is the result of standard regression.
        * PROBABILITY: The output is the probability that the classification is
          correct.
        * MULTIPROBABILITY: The output is an array of probabilities that each
          class is correct ordered by classes seen.
        * RAW: The output is an array of the internal representation of the
          classification process. For example, the raw votes in multi-decision
          tree models.
        * RAW_REGRESSION: The output is an array of the internal representation
          of the regression process. For example, the raw predictions of
          multiple regression trees.
    """
    # pyformat: enable

    return apifunction.ApiFunction.call_(
        self.name() + '.setOutputMode', self, mode
    )

  @staticmethod
  def smileCart(
      # pylint: disable=invalid-name
      maxNodes: _arg_types.Integer | None = None,
      minLeafPopulation: _arg_types.Integer | None = None,
      # pylint: enable=invalid-name
  ) -> Classifier:
    """Returns an empty CART classifier.

    See: Classification and Regression Trees, L. Breiman, J. Friedman, R.
    Olshen, C. Stone Chapman and Hall, 1984.

    Args:
      maxNodes: The maximum number of leaf nodes in each tree. If unspecified,
        defaults to no limit.
      minLeafPopulation: Only create nodes whose training set contains at least
        this many points.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.smileCart', maxNodes, minLeafPopulation
    )

  @staticmethod
  def smileGradientTreeBoost(
      numberOfTrees: _arg_types.Integer,  # pylint: disable=invalid-name
      shrinkage: _arg_types.Number | None = None,
      # pylint: disable=invalid-name
      samplingRate: _arg_types.Number | None = None,
      maxNodes: _arg_types.Integer | None = None,
      # pylint: enable=invalid-name
      loss: _arg_types.String | None = None,
      seed: _arg_types.Integer | None = None,
  ) -> Classifier:
    """Returns an empty Gradient Tree Boost classifier.

    Args:
      numberOfTrees: The number of decision trees to create.
      shrinkage: The shrinkage parameter in (0, 1] controls the learning rate of
        procedure.
      samplingRate: The sampling rate for stochastic tree boosting.
      maxNodes: The maximum number of leaf nodes in each tree. If unspecified,
        defaults to no limit.
      loss: Loss function for regression. One of: LeastSquares,
        LeastAbsoluteDeviation, Huber.
      seed: The randomization seed.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.smileGradientTreeBoost',
        numberOfTrees,
        shrinkage,
        samplingRate,
        maxNodes,
        loss,
        seed,
    )

  @staticmethod
  def smileKNN(
      k: _arg_types.Integer | None = None,
      # pylint: disable-next=invalid-name
      searchMethod: _arg_types.String | None = None,
      metric: _arg_types.String | None = None,
  ) -> Classifier:
    # pyformat: disable
    """Returns an empty k-NN classifier.

    The k-nearest neighbor algorithm (k-NN) is a method for classifying objects
    by a majority vote of its neighbors, with the object being assigned to the
    class most common amongst its k nearest neighbors (k is a positive integer,
    typically small, typically odd).

    Args:
      k: The number of neighbors for classification.
      searchMethod: Search method. The following are valid: [AUTO,
        LINEAR_SEARCH, KD_TREE, COVER_TREE]. AUTO will choose between KD_TREE
        and COVER_TREE depending on the dimension count. Results may vary
        between the different search methods for distance ties and probability
        values. Since performance and results may vary consult with SMILE's
        documentation and other literature.
      metric: The distance metric to use. NOTE: KD_TREE (and AUTO for low
        dimensions) will not use the metric selected. Options are:
        * 'EUCLIDEAN' - Euclidean distance.
        * 'MAHALANOBIS' - Mahalanobis distance.
        * 'MANHATTAN' - Manhattan distance.
        * 'BRAYCURTIS' - Bray-Curtis distance.
    """
    # pyformat: enable

    return apifunction.ApiFunction.call_(
        'Classifier.smileKNN', k, searchMethod, metric
    )

  @staticmethod
  def smileNaiveBayes(
      lambda_: _arg_types.Number | None = None, **kwargs
  ) -> Classifier:
    # pylint: disable=g-doc-args
    """Returns an empty Naive Bayes classifier.

    This classifier assumes that the feature vector consists of positive
    integers, and negative inputs are discarded.

    Args:
      lambda: A smoothing lambda. Used to avoid assigning zero probability to
        classes not seen during training, instead using lambda / (lambda *
        nFeatures).
    """
    # pylint: enable=g-doc-args
    if kwargs:
      if kwargs.keys() != {'lambda'}:
        raise ValueError(
            f'Unexpected arguments: {list(kwargs.keys())}. Expected: lambda.'
        )
      lambda_ = kwargs['lambda']

    return apifunction.ApiFunction.call_('Classifier.smileNaiveBayes', lambda_)

  @staticmethod
  def smileRandomForest(
      # pylint: disable=invalid-name
      numberOfTrees: _arg_types.Integer,
      variablesPerSplit: _arg_types.Integer | None = None,
      minLeafPopulation: _arg_types.Integer | None = None,
      bagFraction: _arg_types.Number | None = None,
      maxNodes: _arg_types.Integer | None = None,
      # pylint: enable=invalid-name
      seed: _arg_types.Integer | None = None,
  ) -> Classifier:
    """Returns an empty Random Forest classifier.

    Args:
      numberOfTrees: The number of decision trees to create.
      variablesPerSplit: The number of variables per split. If unspecified, uses
        the square root of the number of variables.
      minLeafPopulation: Only create nodes whose training set contains at least
        this many points.
      bagFraction: The fraction of input to bag per tree.
      maxNodes: The maximum number of leaf nodes in each tree. If unspecified,
        defaults to no limit.
      seed: The randomization seed.
    """

    return apifunction.ApiFunction.call_(
        'Classifier.smileRandomForest',
        numberOfTrees,
        variablesPerSplit,
        minLeafPopulation,
        bagFraction,
        maxNodes,
        seed,
    )

  @staticmethod
  def spectralRegion(
      coordinates: _arg_types.List, schema: _arg_types.List | None = None
  ) -> Classifier:
    """Returns a spectral region classifier.

    Creates a classifier that tests if its inputs lie within a polygon defined
    by a set  of coordinates in an arbitrary 2D coordinate system. Each input to
    be classified  must have 2 values (e.g., images must have 2 bands). The
    result will be 1 wherever  the input values are contained within the given
    polygon and 0 otherwise.

    Args:
      coordinates: The coordinates of the polygon, as a list of rings. Each ring
        is a list of coordinate pairs (e.g., [u1, v1, u2, v2, ..., uN, vN]). No
        edge may intersect any other edge. The resulting classification will be
        a 1 wherever the input values are within the interior of the given
        polygon, that is, an odd number of polygon edges must be crossed to get
        outside the polygon and 0 otherwise.
      schema: The classifier's schema. A list of band or property names that the
        classifier will operate on. Since this classifier doesn't undergo a
        training step, these have to be specified manually. Defaults to ['u',
        'v'].
    """

    return apifunction.ApiFunction.call_(
        'Classifier.spectralRegion', coordinates, schema
    )

  def train(
      # classifier: _ClassifierType,
      self,
      features: _arg_types.FeatureCollection,
      # pylint: disable=invalid-name
      classProperty: _arg_types.String,
      inputProperties: _arg_types.List | None = None,
      # pylint: enable=invalid-name
      subsampling: _arg_types.Number | None = None,
      # pylint: disable-next=invalid-name
      subsamplingSeed: _arg_types.Integer | None = None,
  ) -> Classifier:
    """Returns a trained classifier.

    Trains the classifier on a collection of features, using the specified
    numeric properties of each feature as training data. The geometry of the
    features is ignored.

    Args:
      features: The collection to train on.
      classProperty: The name of the property containing the class value. Each
        feature must have this property and its value must be numeric.
      inputProperties: The list of property names to include as training data.
        Each feature must have all these properties and their values must be
        numeric. This argument is optional if the input collection contains a
        'band_order' property, (as produced by Image.sample).
      subsampling: An optional subsampling factor, within (0, 1].
      subsamplingSeed: A randomization seed to use for subsampling.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.train',
        self,
        features,
        classProperty,
        inputProperties,
        subsampling,
        subsamplingSeed,
    )



#!/usr/bin/env python3
"""Test for ee._cloud_api_utils."""

import json
from unittest import mock
import warnings

from google.oauth2 import credentials
from googleapiclient import discovery
import requests

import unittest
from ee import _cloud_api_utils
from ee import ee_exception


class CloudApiUtilsTest(unittest.TestCase):

  def test_build_cloud_resource(self):
    base = 'https://earthengine.basetest'
    path = '$discovery/rest?version=v1&prettyPrint=false'
    def check_build(api, unused_version, **kwargs):
      self.assertEqual('earthengine', api)
      self.assertEqual(base + '/' + path, kwargs.get('discoveryServiceUrl'))
      http = kwargs.get('http')
      self.assertIsNotNone(http)
      self.assertIsNone(http.credentials.quota_project_id)
      return mock.Mock()

    with mock.patch.object(discovery, 'build', side_effect=check_build) as run:
      session = requests.Session()
      cred_args = dict(token=None, refresh_token='rt', quota_project_id='qp')
      resource = _cloud_api_utils.build_cloud_resource(
          base, session, credentials=credentials.Credentials(**cred_args)
      )
      self.assertEqual(base, resource._baseUrl)
      run.assert_called_once()

  def test_convert_dict_simple(self):
    result = _cloud_api_utils._convert_dict({
        'x': 99,
        'y': 2
    }, {
        'x': 'a',
        'y': ('c', lambda x: x + 1)
    })
    self.assertEqual({'a': 99, 'c': 3}, result)

  def test_convert_dict_with_defaults(self):
    result = _cloud_api_utils._convert_dict(
        {
            'x': 99
        }, {'x': 'a'}, defaults={
            'q': 'r',
            'a': 'aaa'
        })
    self.assertEqual({'a': 99, 'q': 'r'}, result)

  def test_convert_dict_with_warnings(self):
    with warnings.catch_warnings(record=True) as w:
      result = _cloud_api_utils._convert_dict({
          'x': 1,
          'y': 2
      }, {
          'x': 'flan',
          'y': 'flan'
      })
      self.assertEqual({'flan': 2}, result)
      self.assertEqual('Multiple request parameters converted to flan',
                       str(w[0].message))

    with warnings.catch_warnings(record=True) as w:
      result = _cloud_api_utils._convert_dict({'x': 1, 'y': 2}, {'x': 'flan'})
      self.assertEqual({'flan': 1}, result)

    with warnings.catch_warnings(record=True) as w:
      result = _cloud_api_utils._convert_dict(
          {
              'x': 1,
              'y': 2
          }, {'x': 'flan'}, key_warnings=True)
      self.assertEqual({'flan': 1}, result)
      self.assertEqual('Unrecognized key y ignored', str(w[0].message))

  def test_convert_value(self):
    self.assertEqual('x', _cloud_api_utils._convert_value('a', {'a': 'x'}, 'z'))
    self.assertEqual('z', _cloud_api_utils._convert_value('b', {'a': 'x'}, 'z'))

  def test_convert_msec_to_timestamp(self):
    self.assertEqual('2018-07-31T20:46:28.888000Z',
                     _cloud_api_utils._convert_msec_to_timestamp(1533069988888))

  def test_convert_timestamp_to_msec(self):
    self.assertEqual(
        1533069988888,
        _cloud_api_utils._convert_timestamp_to_msec(
            '2018-07-31T20:46:28.888000Z'))
    self.assertEqual(
        1533069988000,
        _cloud_api_utils._convert_timestamp_to_msec(
            '2018-07-31T20:46:28Z'))

  def test_convert_bounding_box_to_geo_json(self):
    geo_json = _cloud_api_utils._convert_bounding_box_to_geo_json(
        [-105, 45, 10, 60])
    geo_json_parsed = json.loads(geo_json)
    self.assertEqual({
        'type':
            'Polygon',
        'coordinates': [[[-105, 45], [10, 45], [10, 60], [-105, 60], [-105, 45]]
                       ]
    }, geo_json_parsed)

  def test_convert_asset_type_for_create_asset(self):
    self.assertEqual(
        'IMAGE_COLLECTION',
        _cloud_api_utils.convert_asset_type_for_create_asset('ImageCollection'))
    self.assertEqual(
        'IMAGE_COLLECTION',
        _cloud_api_utils.convert_asset_type_for_create_asset(
            'IMAGE_COLLECTION'))
    self.assertEqual(
        'FOLDER',
        _cloud_api_utils.convert_asset_type_for_create_asset('Folder'))
    self.assertEqual(
        'FOLDER',
        _cloud_api_utils.convert_asset_type_for_create_asset('FOLDER'))

  def test_convert_asset_id_to_asset_name(self):
    self.assertEqual(
        'projects/earthengine-public/assets/path/to/asset',
        _cloud_api_utils.convert_asset_id_to_asset_name('path/to/asset'))
    self.assertEqual(
        'projects/earthengine-legacy/assets/users/path/to/asset',
        _cloud_api_utils.convert_asset_id_to_asset_name('users/path/to/asset'))
    self.assertEqual(
        'projects/earthengine-legacy/assets/projects/path/to/asset',
        _cloud_api_utils.convert_asset_id_to_asset_name(
            'projects/path/to/asset'))
    self.assertEqual(
        'projects/foobar/assets/baz',
        _cloud_api_utils.convert_asset_id_to_asset_name(
            'projects/foobar/assets/baz'))

  def test_split_asset_name(self):
    parent, asset_id = _cloud_api_utils.split_asset_name(
        'projects/earthengine-legacy/assets/users/path/to/asset')
    self.assertEqual('projects/earthengine-legacy', parent)
    self.assertEqual('users/path/to/asset', asset_id)

  def test_convert_operation_name_to_task_id(self):
    self.assertEqual(
        'taskId',
        _cloud_api_utils.convert_operation_name_to_task_id(
            'operations/taskId'))
    self.assertEqual(
        'taskId',
        _cloud_api_utils.convert_operation_name_to_task_id(
            'projects/test/operations/taskId'))
    self.assertEqual(
        'taskId',
        _cloud_api_utils.convert_operation_name_to_task_id(
            'projects/operations/operations/taskId'))
    self.assertEqual(
        'taskId', _cloud_api_utils.convert_operation_name_to_task_id('taskId'))

  def test_convert_task_id_to_operation_name(self):
    self.assertEqual(
        'projects/earthengine-legacy/operations/taskId',
        _cloud_api_utils.convert_task_id_to_operation_name(
            'earthengine-legacy', 'taskId'
        ),
    )

  def test_encode_number_as_cloud_value(self):
    self.assertEqual({
        'constantValue': 112233
    }, _cloud_api_utils.encode_number_as_cloud_value(112233))
    # Large integer not representable as double.
    self.assertEqual({
        'integerValue': '112233445566778899'
    }, _cloud_api_utils.encode_number_as_cloud_value(112233445566778899))
    # Large integer representable as double.
    self.assertEqual({
        'constantValue': 3.402823669209385e+38
    }, _cloud_api_utils.encode_number_as_cloud_value(1 << 128))

  def test_convert_algorithms(self):
    result = _cloud_api_utils.convert_algorithms({
        'algorithms': [
            {
                'name': 'algorithms/algNoArgs',
                'description': 'desc',
                'returnType': 'ret'
            },
            {
                'name': 'algorithms/algNoDesc',
                'arguments': [
                    {
                        'argumentName': 'argNoDesc'
                    },
                    {
                        'argumentName': 'argOptional',
                        'description': 'descArg',
                        'type': 't',
                        'optional': True,
                        'defaultValue': [1, 2]
                    }]
            },
            {
                'name': 'algorithms/algHidden',
                'description': 'desc',
                'returnType': 'ret',
                'hidden': True
            },
            {
                'name': 'algorithms/algPreview',
                'description': 'desc',
                'returnType': 'ret',
                'preview': True
            },
            {
                'name': 'algorithms/algDeprecated',
                'description': 'desc',
                'returnType': 'ret',
                'deprecated': True,
                'deprecationReason': 'reason'
            }]
    })
    self.assertEqual({
        'algNoArgs': {
            'description': 'desc',
            'returns': 'ret',
            'args': []
        },
        'algNoDesc': {
            'description': '',
            'returns': '',
            'args': [
                {
                    'name': 'argNoDesc',
                    'type': '',
                    'description': ''
                },
                {
                    'name': 'argOptional',
                    'description': 'descArg',
                    'type': 't',
                    'optional': True,
                    'default': [1, 2]
                }]
        },
        'algHidden': {
            'description': 'desc',
            'returns': 'ret',
            'args': [],
            'hidden': True
        },
        'algPreview': {
            'description': 'desc',
            'returns': 'ret',
            'args': [],
            'preview': True
        },
        'algDeprecated': {
            'description': 'desc',
            'returns': 'ret',
            'args': [],
            'deprecated': 'reason'
        }
    }, result)

  def test_convert_empty_algorithms(self):
    result = _cloud_api_utils.convert_algorithms({})
    self.assertEqual({}, result)

  def test_convert_to_image_file_format(self):
    self.assertEqual('AUTO_JPEG_PNG',
                     _cloud_api_utils.convert_to_image_file_format(None))
    self.assertEqual('AUTO_JPEG_PNG',
                     _cloud_api_utils.convert_to_image_file_format('auto'))
    self.assertEqual('JPEG',
                     _cloud_api_utils.convert_to_image_file_format('jpg'))
    self.assertEqual('JPEG',
                     _cloud_api_utils.convert_to_image_file_format('jpeg'))
    self.assertEqual('PNG',
                     _cloud_api_utils.convert_to_image_file_format('png'))
    self.assertEqual('GEO_TIFF',
                     _cloud_api_utils.convert_to_image_file_format('GeoTIFF'))
    self.assertEqual('TF_RECORD_IMAGE',
                     _cloud_api_utils.convert_to_image_file_format('TFRecord'))

  def test_convert_to_table_file_format(self):
    self.assertEqual('CSV',
                     _cloud_api_utils.convert_to_table_file_format('csv'))
    self.assertEqual('GEO_JSON',
                     _cloud_api_utils.convert_to_table_file_format('GeoJSON'))
    self.assertEqual('TF_RECORD_TABLE',
                     _cloud_api_utils.convert_to_table_file_format('TFRecord'))

  def test_convert_to_band_list(self):
    self.assertEqual([], _cloud_api_utils.convert_to_band_list(None))
    self.assertEqual(['a'], _cloud_api_utils.convert_to_band_list('a'))
    self.assertEqual(['a', 'b'], _cloud_api_utils.convert_to_band_list('a,b'))
    self.assertEqual(['a', 'b'],
                     _cloud_api_utils.convert_to_band_list(['a', 'b']))

  def test_convert_to_visualization_options(self):
    with self.assertRaises(ee_exception.EEException):
      _cloud_api_utils.convert_to_visualization_options({
          'min': '0,1',
          'max': '2'
      })
    with self.assertRaises(ee_exception.EEException):
      _cloud_api_utils.convert_to_visualization_options({
          'gain': '0,1',
          'bias': '2'
      })
    with self.assertRaises(ee_exception.EEException):
      _cloud_api_utils.convert_to_visualization_options({
          'gain': '0,1',
          'min': '0,1'
      })
    with self.assertRaises(ee_exception.EEException):
      _cloud_api_utils.convert_to_visualization_options({
          'gamma': '2,3'
      })
    with self.assertRaises(ee_exception.EEException):
      _cloud_api_utils.convert_to_visualization_options({'gamma': '0,1'})
    self.assertEqual({
        'ranges': [{'min': 0, 'max': 1}, {'min': 0, 'max': 2}],
        'gamma': {'value': 2.2}
    },
                     _cloud_api_utils.convert_to_visualization_options({
                         'max': '1,2',
                         'gamma': '2.2'
                     }))
    self.assertEqual({
        'ranges': [{'min': 0, 'max': 1}, {'min': -.5, 'max': 0}],
        'paletteColors': ['0', '1', '2']
    },
                     _cloud_api_utils.convert_to_visualization_options({
                         'gain': '2,4',
                         'bias': '0,2',
                         'palette': '0,1,2',
                     }))

  def test_convert_operation_to_task(self):
    self.assertEqual(
        {
            'start_timestamp_ms': 1538676004000,
            'state': 'COMPLETED',
            'description': 'Ingest image: "an/image"',
            'creation_timestamp_ms': 1538676001749,
            'id': '7T42Q7FH4KSIXQKGT6MJFBPX',
            'update_timestamp_ms': 1538676053218,
            'task_type': 'INGEST_IMAGE',
            'destination_uris': ['https://test.com'],
            'name': 'projects/test/operations/7T42Q7FH4KSIXQKGT6MJFBPX',
            'attempt': 42,
            'priority': 100,
        },
        _cloud_api_utils.convert_operation_to_task({
            'metadata': {
                'createTime': '2018-10-04T18:00:01.749999Z',
                'updateTime': '2018-10-04T18:00:53.218488Z',
                'description': 'Ingest image: "an/image"',
                'startTime': '2018-10-04T18:00:04Z',
                'state': 'SUCCEEDED',
                'endTime': '2018-10-04T18:00:53.218488Z',
                'type': 'INGEST_IMAGE',
                'destinationUris': ['https://test.com'],
                'attempt': 42,
                'priority': 100,
            },
            'done': True,
            'name': 'projects/test/operations/7T42Q7FH4KSIXQKGT6MJFBPX',
        }))

  def test_convert_iam_policy_to_acl(self):
    self.assertEqual({
        'owners': ['user:owner@owner.domain'],
        'writers': [],
        'readers': ['group:readerGroup', 'user:readerUser'],
        'all_users_can_read': True
    }, _cloud_api_utils.convert_iam_policy_to_acl({
        'bindings': [{
            'role': 'roles/owner',
            'members': ['user:owner@owner.domain']
        }, {
            'role': 'roles/viewer',
            'members': ['group:readerGroup', 'allUsers', 'user:readerUser']
        }]
    }))

  def test_convert_acl_to_iam_policy(self):
    self.assertEqual({
        'bindings': [{
            'role': 'roles/owner',
            'members': ['user:owner@owner.domain']
        }, {
            'role': 'roles/viewer',
            'members': ['group:readerGroup', 'user:readerUser', 'allUsers']
        }]
    }, _cloud_api_utils.convert_acl_to_iam_policy({
        'owners': ['user:owner@owner.domain'],
        'writers': [],
        'readers': ['group:readerGroup', 'user:readerUser'],
        'all_users_can_read': True
    }))

  def test_convert_to_grid_dimension(self):
    self.assertEqual({'width': 123, 'height': 123},
                     _cloud_api_utils.convert_to_grid_dimensions(123))
    self.assertEqual({'width': 123, 'height': 123},
                     _cloud_api_utils.convert_to_grid_dimensions(123))
    self.assertEqual({'width': 123, 'height': 234},
                     _cloud_api_utils.convert_to_grid_dimensions((123, 234)))

  def test_to_image_old_source(self):
    """Converts legacy keys in the source object."""
    old_sources = [{
        'primaryPath': 'path1',
        'additionalPaths': ['path2'],
        'affineTransform': {
            'scaleX': 1,
            'shearX': 2,
            'translateX': 3,
            'shearY': 4,
            'scaleY': 5,
            'translateY': 6,
        }
    }]
    expected = [{
        'uris': ['path1', 'path2'],
        'affineTransform': {
            'scaleX': 1,
            'shearX': 2,
            'translateX': 3,
            'shearY': 4,
            'scaleY': 5,
            'translateY': 6,
        }
    }]
    self.assertEqual(
        expected,
        _cloud_api_utils.convert_sources_to_one_platform_sources(old_sources))

  def test_to_image_one_platform_source(self):
    """Leaves new keys in the source object unchanged."""
    old_sources = [{
        'uris': ['path1'],
        'affineTransform': {
            'scaleX': 1,
            'shearX': 2,
            'translateX': 3,
            'shearY': 4,
            'scaleY': 5,
            'translateY': 6,
        }
    }]
    expected = [{
        'uris': ['path1'],
        'affineTransform': {
            'scaleX': 1,
            'shearX': 2,
            'translateX': 3,
            'shearY': 4,
            'scaleY': 5,
            'translateY': 6,
        }
    }]
    self.assertEqual(
        expected,
        _cloud_api_utils.convert_sources_to_one_platform_sources(old_sources))

  def test_to_shape_one_platform_source(self):
    old_sources = [{
        'charset':
            'UTF-8',
        'maxError':
            1,
        'maxVertices':
            100000,
        'primaryPath':
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.shp',
        'additionalPaths': [
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.dbf',
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.prj',
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.shx',
        ],
    }]
    expected = [{
        'charset':
            'UTF-8',
        'maxErrorMeters':
            1,
        'maxVertices':
            100000,
        'uris': [
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.shp',
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.dbf',
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.prj',
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/test.shx',
        ],
    }]
    self.assertEqual(
        expected,
        _cloud_api_utils.convert_sources_to_one_platform_sources(old_sources))

  def test_to_csv_one_platform_source(self):
    old_sources = [{
        'charset':
            'UTF-8',
        'maxError':
            1,
        'maxVertices':
            1000000,
        'geodesic':
            True,
        'primaryGeometryColumn':
            'geometry0',
        'xColumn':
            'x',
        'yColumn':
            'y',
        'primaryPath':
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/wildlife.csv',
    }]
    expected = [{
        'charset':
            'UTF-8',
        'maxErrorMeters':
            1,
        'maxVertices':
            1000000,
        'geodesic':
            True,
        'primaryGeometryColumn':
            'geometry0',
        'xColumn':
            'x',
        'yColumn':
            'y',
        'uris': [
            'gs://ee.google.com.a.appspot.com/qGc_ZVNWLKpokgLv/wildlife.csv',
        ],
    }]
    self.assertEqual(
        expected,
        _cloud_api_utils.convert_sources_to_one_platform_sources(old_sources))

  def test_convert_to_operation_state(self):
    def convert(state):
      return _cloud_api_utils.convert_to_operation_state(state)

    self.assertEqual('CANCELLED', convert('CANCELLED'))
    self.assertEqual('CANCELLING', convert('CANCEL_REQUESTED'))
    self.assertEqual('CANCELLING', convert('CANCELLING'))
    self.assertEqual('FAILED', convert('FAILED'))
    self.assertEqual('PENDING', convert('PENDING'))
    self.assertEqual('PENDING', convert('READY'))
    self.assertEqual('RUNNING', convert('RUNNING'))
    self.assertEqual('SUCCEEDED', convert('COMPLETED'))
    self.assertEqual('SUCCEEDED', convert('SUCCEEDED'))
    self.assertEqual('UNKNOWN', convert('random_string'))
    self.assertEqual('UNKNOWN', convert('UNKNOWN'))


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Test for the ee.deserializer module."""

import json

import unittest
import ee
from ee import apitestcase
from ee import deserializer
from ee import serializer


class DeserializerTest(apitestcase.ApiTestCase):

  def test_legacy_roundtrip(self):
    """Verifies a round trip of a comprehensive serialization case."""
    encoded = apitestcase.ENCODED_JSON_SAMPLE
    decoded = deserializer.decode(encoded)
    re_encoded = json.loads(serializer.toJSON(decoded, for_cloud_api=False))
    self.assertEqual(encoded, re_encoded)
    # Round-trip the decoded object through the Cloud API
    cloud_encoded = json.loads(serializer.toJSON(decoded, for_cloud_api=True))
    cloud_decoded = deserializer.decode(cloud_encoded)  # Supports both formats
    cloud_re_encoded = json.loads(
        serializer.toJSON(cloud_decoded, for_cloud_api=True))
    self.assertEqual(cloud_encoded, cloud_re_encoded)

  def test_cloud_roundtrip(self):
    """Verifies a round trip of a comprehensive serialization case."""
    cloud_encoded = apitestcase.ENCODED_CLOUD_API_JSON_SAMPLE
    cloud_decoded = deserializer.decode(cloud_encoded)  # Supports both formats
    cloud_re_encoded = json.loads(
        serializer.toJSON(cloud_decoded, for_cloud_api=True))
    self.assertEqual(cloud_encoded, cloud_re_encoded)
    # Round-trip the decoded object through the legacy API
    encoded = json.loads(serializer.toJSON(cloud_decoded, for_cloud_api=False))
    decoded = deserializer.decode(encoded)
    re_encoded = json.loads(serializer.toJSON(decoded, for_cloud_api=False))
    self.assertEqual(encoded, re_encoded)

  def test_cast(self):
    """Verifies that decoding casts the result to the right class."""
    input_image = ee.Image(13).addBands(42)
    output = deserializer.fromJSON(serializer.toJSON(input_image))
    self.assertIsInstance(output, ee.Image)
    cloud_output = deserializer.fromCloudApiJSON(
        serializer.toJSON(input_image, for_cloud_api=True))
    self.assertIsInstance(cloud_output, ee.Image)

  def test_reuse(self):
    """Verifies that decoding results can be used and re-encoded."""
    input_image = ee.Image(13)
    output = deserializer.fromJSON(serializer.toJSON(input_image))
    self.assertEqual(
        output.addBands(42).serialize(),
        input_image.addBands(42).serialize())
    cloud_output = deserializer.fromCloudApiJSON(
        serializer.toJSON(input_image, for_cloud_api=True))
    self.assertEqual(
        cloud_output.addBands(42).serialize(),
        input_image.addBands(42).serialize())

  def test_image_expression(self):
    """Verifies that ee.Image.expression results can be re-encoded."""
    image = ee.Image(13)
    expression = image.expression('x', {'x': image})
    expression_encoded = serializer.toJSON(expression)
    expression_decoded = deserializer.decode(json.loads(expression_encoded))
    expression_re_encoded = serializer.toJSON(expression_decoded)
    self.assertEqual(expression_encoded, expression_re_encoded)

  def test_unknown_value_ref(self):
    """Verifies raising Unknown ValueRef in _decodeValue()."""
    encoded = {
        'type': 'CompoundValue',
        'scope': [['key', {'type': 'ValueRef', 'value': 'bar'}]],
    }
    with self.assertRaisesRegex(
        ee.EEException, "Unknown ValueRef: {'type': 'ValueRef', 'value': 'bar'}"
    ):
      deserializer.decode(encoded)

  def test_duplicate_scope_key(self):
    """Verifies raising duplicate scope key in decode()."""
    encoded = {
        'type': 'CompoundValue',
        'scope': [['a', 1], ['a', 2]],
        'value': 3,
    }
    with self.assertRaisesRegex(
        ee.EEException, 'Duplicate scope key "a" in scope #1.'
    ):
      deserializer.decode(encoded)

  def test_cannot_decode_object(self):
    """Verifies raising EEException for non-dict objects in _decodeValue()."""
    with self.assertRaisesRegex(
        ee.EEException, r'Cannot decode object: \(1\+1j\)'
    ):
      deserializer.decode([1 + 1j])

  def test_invalid_date_value(self):
    """Verifies EEException for invalid date values in _decodeValue()."""
    with self.assertRaisesRegex(
        ee.EEException, r'Invalid date value: not-a-number'
    ):
      deserializer.decode({'type': 'Date', 'value': 'not-a-number'})

  def test_invocation_of_custom_function(self):
    """Verifies decoding of an Invocation of a CustomFunction."""
    encoded = {
        'type': 'Invocation',
        'function': {
            'type': 'Function',
            'argumentNames': ['foo'],
            'body': {'type': 'ArgumentRef', 'value': 'foo'},
        },
        'arguments': {'foo': 1},
    }
    decoded = deserializer.decode(encoded)
    self.assertIsInstance(decoded, ee.ComputedObject)

  def test_date_decode(self):
    """Verifies decoding of a Date."""
    encoded = {'type': 'Date', 'value': 1609459200000000}
    decoded = deserializer.decode(encoded)
    self.assertIsInstance(decoded, ee.Date)
    self.assertEqual({'value': 1609459200000}, decoded.args)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Models."""

from ee import _arg_types
from ee import apifunction
from ee import computedobject
from ee import deprecation
from ee import featurecollection
from ee import image as ee_image


class Model(computedobject.ComputedObject):
  """An object to represent an Earth Engine Model.

  Example:
    model = ee.Model.fromVertexAi(
        endpoint='endpoint-name',
        inputTileSize=[8, 8],
        outputBands={
            'probability': {'type': ee.PixelType.float(), 'dimensions': 1}
        },
    )

  Please visit one of the following links for more info:
    - https://developers.google.com/earth-engine/guides/machine-learning
    - https://developers.google.com/earth-engine/guides/tensorflow-vertex
  """

  _initialized: bool = False

  def __init__(self, model: computedobject.ComputedObject):
    """Creates a Model wrapper.

    Args:
      model: A Model to cast.
    """
    self.initialize()

    if isinstance(model, computedobject.ComputedObject):
      # There is no server-side constructor for ee.Model. Pass the object as-is
      # to the server in case it is intended to be a Model cast.
      super().__init__(model.func, model.args, model.varName)
      return

    raise TypeError('Model constructor can only cast to Model.')

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Model'

  @staticmethod
  @deprecation.Deprecated('Migrate to Vertex AI')
  def fromAiPlatformPredictor(
      # pylint: disable=invalid-name
      projectName: _arg_types.Any | None = None,
      projectId: _arg_types.String | None = None,
      modelName: _arg_types.String | None = None,
      # pylint: enable=invalid-name
      version: _arg_types.String | None = None,
      region: _arg_types.String | None = None,
      # pylint: disable=invalid-name
      inputProperties: _arg_types.List | None = None,
      inputTypeOverride: _arg_types.Dictionary | None = None,
      inputShapes: _arg_types.Dictionary | None = None,
      # pylint: enable=invalid-name
      proj: _arg_types.Projection | None = None,
      # pylint: disable=invalid-name
      fixInputProj: _arg_types.Bool | None = None,
      inputTileSize: _arg_types.List | None = None,
      inputOverlapSize: _arg_types.List | None = None,
      outputTileSize: _arg_types.List | None = None,
      outputBands: _arg_types.Dictionary | None = None,
      outputProperties: _arg_types.Dictionary | None = None,
      outputMultiplier: _arg_types.Number | None = None,
      # pylint: enable=invalid-name
  ) -> 'Model':
    """Returns an ee.Model from a description of an AI Platform prediction model.

    DEPRECATED: Cloud AI Platform is deprecated migrate to Vertex AI:
    https://developers.google.com/earth-engine/guides/ee-vertex-migrate

    (See https://cloud.google.com/ml-engine/).

    Args:
      projectName: The Google Cloud project that owns the model. Deprecated: use
        "projectId" instead.
      projectId: The ID of the Google Cloud project that owns the model.
      modelName: The name of the model.
      version: The model version. Defaults to the AI Platform default model
        version.
      region: The model deployment region. Defaults to "us-central1".
      inputProperties: Properties passed with each prediction instance. Image
        predictions are tiled, so these properties will be replicated into each
        image tile instance. Defaults to no properties.
      inputTypeOverride: Types to which model inputs will be coerced if
        specified. Both Image bands and Image/Feature properties are valid.
      inputShapes: The fixed shape of input array bands. For each array band not
        specified, the fixed array shape will be automatically deduced from a
        non-masked pixel.
      proj: The input projection at which to sample all bands. Defaults to the
        default projection of an image's first band.
      fixInputProj: If true, pixels will be sampled in a fixed projection
        specified by 'proj'. The output projection is used otherwise. Defaults
        to false.
      inputTileSize: Rectangular dimensions of pixel tiles passed in to
        prediction instances. Required for image predictions.
      inputOverlapSize: Amount of adjacent-tile overlap in X/Y along each edge
        of pixel tiles passed in to prediction instances. Defaults to [0, 0].
      outputTileSize: Rectangular dimensions of pixel tiles returned from AI
        Platform. Defaults to the value in 'inputTileSize'.
      outputBands: A map from output band names to a dictionary of output band
        info. Valid band info fields are 'type' and 'dimensions'. 'type' should
        be a ee.PixelType describing the output band, and 'dimensions' is an
        optional integer with the number of dimensions in that band e.g.,
        "outputBands: {'p': {'type': ee.PixelType.int8(), 'dimensions': 1}}".
        Required for image predictions.
      outputProperties: A map from output property names to a dictionary of
        output property info. Valid property info fields are 'type' and
        'dimensions'. 'type' should be a ee.PixelType describing the output
        property, and 'dimensions' is an optional integer with the number of
        dimensions for that property if it is an array e.g., "outputBands: {'p':
        {'type': ee.PixelType.int8(), 'dimensions': 1}}". Required for
        predictions from FeatureCollections.
      outputMultiplier: An approximation to the increase in data volume for the
        model outputs over the model inputs. If specified this must be >= 1.
        This is only needed if the model produces more data than it consumes,
        e.g., a model that takes 5 bands and produces 10 outputs per pixel.

    Returns:
      An ee.Model.
    """

    return apifunction.ApiFunction.call_(
        'Model.fromAiPlatformPredictor',
        projectName,
        projectId,
        modelName,
        version,
        region,
        inputProperties,
        inputTypeOverride,
        inputShapes,
        proj,
        fixInputProj,
        inputTileSize,
        inputOverlapSize,
        outputTileSize,
        outputBands,
        outputProperties,
        outputMultiplier,
    )

  @staticmethod
  def fromVertexAi(
      endpoint: _arg_types.String,
      # pylint: disable=invalid-name
      inputProperties: _arg_types.List | None = None,
      inputTypeOverride: _arg_types.Dictionary | None = None,
      inputShapes: _arg_types.Dictionary | None = None,
      # pylint: enable=invalid-name
      proj: _arg_types.Projection | None = None,
      # pylint: disable=invalid-name
      fixInputProj: _arg_types.Bool | None = None,
      inputTileSize: _arg_types.List | None = None,
      inputOverlapSize: _arg_types.List | None = None,
      outputTileSize: _arg_types.List | None = None,
      outputBands: _arg_types.Dictionary | None = None,
      outputProperties: _arg_types.Dictionary | None = None,
      outputMultiplier: _arg_types.Number | None = None,
      maxPayloadBytes: _arg_types.Integer | None = None,
      payloadFormat: _arg_types.String | None = None,
      # pylint: enable=invalid-name
  ) -> 'Model':
    """Returns an ee.Model from a description of a Vertex AI model endpoint.

    See https://cloud.google.com/vertex-ai.

    Warning: This method is in public preview and may undergo breaking changes.

    Args:
      endpoint: The endpoint name for predictions.
      inputProperties: Properties passed with each prediction instance. Image
        predictions are tiled, so these properties will be replicated into each
        image tile instance. Defaults to no properties.
      inputTypeOverride: Types to which model inputs will be coerced if
        specified. Both Image bands and Image/Feature properties are valid.
      inputShapes: The fixed shape of input array bands. For each array band not
        specified, the fixed array shape will be automatically deduced from a
        non-masked pixel.
      proj: The input projection at which to sample all bands. Defaults to the
        default projection of an image's first band.
      fixInputProj: If true, pixels will be sampled in a fixed projection
        specified by 'proj'. The output projection is used otherwise. Defaults
        to false.
      inputTileSize: Rectangular dimensions of pixel tiles passed in to
        prediction instances. Required for image predictions.
      inputOverlapSize: Amount of adjacent-tile overlap in X/Y along each edge
        of pixel tiles passed in to prediction instances. Defaults to [0, 0].
      outputTileSize: Rectangular dimensions of pixel tiles returned from AI
        Platform. Defaults to the value in 'inputTileSize'.
      outputBands: A map from output band names to a dictionary of output band
        info. Valid band info fields are 'type' and 'dimensions'. 'type' should
        be a ee.PixelType describing the output band, and 'dimensions' is an
        optional integer with the number of dimensions in that band e.g.,
        "outputBands: {'p': {'type': ee.PixelType.int8(), 'dimensions': 1}}".
        Required for image predictions.
      outputProperties: A map from output property names to a dictionary of
        output property info. Valid property info fields are 'type' and
        'dimensions'. 'type' should be a ee.PixelType describing the output
        property, and 'dimensions' is an optional integer with the number of
        dimensions for that property if it is an array e.g., "outputBands: {'p':
        {'type': ee.PixelType.int8(), 'dimensions': 1}}". Required for
        predictions from FeatureCollections.
      outputMultiplier: An approximation to the increase in data volume for the
        model outputs over the model inputs. If specified this must be >= 1.
        This is only needed if the model produces more data than it consumes,
        e.g., a model that takes 5 bands and produces 10 outputs per pixel.
      maxPayloadBytes: The prediction payload size limit in bytes. Defaults to
        1.5MB (1500000 bytes.)
      payloadFormat: The payload format of entries in prediction requests and
        responses. One of: ['SERIALIZED_TF_TENSORS, 'RAW_JSON', 'ND_ARRAYS',
        'GRPC_TF_TENSORS', 'GRPC_SERIALIZED_TF_TENSORS', 'GRPC_TF_EXAMPLES'].
        Defaults to 'SERIALIZED_TF_TENSORS'.

    Returns:
      An ee.Model.
    """

    return apifunction.ApiFunction.call_(
        'Model.fromVertexAi',
        endpoint,
        inputProperties,
        inputTypeOverride,
        inputShapes,
        proj,
        fixInputProj,
        inputTileSize,
        inputOverlapSize,
        outputTileSize,
        outputBands,
        outputProperties,
        outputMultiplier,
        maxPayloadBytes,
        payloadFormat,
    )

  def predictImage(self, image: _arg_types.Image) -> ee_image.Image:
    """Returns an image with predictions from pixel tiles of an image.

    The predictions are merged as bands with the input image.

    The model will receive 0s in place of masked pixels. The masks of predicted
    output bands are the minimum of the masks of the inputs.

    Args:
      image: The input image.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.predictImage', self, image
    )

  def predictProperties(
      self, collection: _arg_types.FeatureCollection
  ) -> featurecollection.FeatureCollection:
    """Returns a feature collection with predictions for each feature.

    Predicted properties are merged with the properties of the input feature.

    Args:
      collection: The input collection.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.predictProperties', self, collection
    )
